# -*- coding: utf-8 -*-
# (c) 2015-2016 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import os
import re
from collections import OrderedDict
from cornice.util import to_list
from pprint import pprint
from binascii import hexlify
from tabulate import tabulate
from appdirs import user_cache_dir
from ctypes import c_uint8, c_uint16, c_uint32, c_int8, c_int16, c_int32
from pyclibrary.c_parser import CParser
from pyclibrary import CLibrary, auto_init
from pyclibrary.utils import add_header_locations, add_library_locations
from sympy.core.sympify import sympify
from twisted.logger import Logger
from kotori.daq.intercom.pyclibrary_ext.c_parser import CParserEnhanced
from kotori.daq.intercom.pyclibrary_ext.backend_ctypes import monkeypatch_pyclibrary_ctypes_struct
from kotori.util.common import slm

logger = Logger()

class LibraryAdapter(object):

    def __init__(self, header_files, library_file, include_path=None, library_path=None, cache_path=None):
        self.header_files = to_list(header_files)
        self.library_file = library_file
        self.include_path = include_path or os.curdir
        self.library_path = library_path or os.curdir
        self.cache_path = cache_path or './var'

        self.include_path = os.path.abspath(self.include_path)
        self.library_path = os.path.abspath(self.library_path)

        cache_key = \
            self.include_path.replace('/', '_') + \
            u'-' + \
            u'_'.join(self.header_files) + \
            u'.pyclibrary'
        self.cache_file = os.path.join(self.cache_path, cache_key)

        logger.info('Setting up library "{}" with headers "{}", cache file is "{}"'.format(
            self.library_file, ', '.join(self.header_files), self.cache_file))

        # holding the library essentials
        self.parser = None
        self.clib = None
        self.annotations = None

        self.setup()
        self.parse()
        self.parse_annotations()
        self.load()

    def setup(self):

        # counter "ValueError: number of bits invalid for bit field"
        monkeypatch_pyclibrary_ctypes_struct()

        # register header- and library paths
        # https://pyclibrary.readthedocs.org/en/latest/get_started/configuration.html#specifying-headers-and-libraries-locations
        # TODO: this probably acts on a global basis; think about it
        if self.include_path:
            add_header_locations([self.include_path])
        if self.library_path:
            add_library_locations([self.library_path])

        # define extra types suitable for embedded use
        types = {
            'uint8_t': c_uint8,
            'uint16_t': c_uint16,
            'uint32_t': c_uint32,
            'int8_t': c_int8,
            'int16_t': c_int16,
            'int32_t': c_int32,
            }

        # TODO: this probably acts on a global basis; think about it
        if not (CParser._init or CLibrary._init):
            auto_init(extra_types=types)

    def parse(self):
        # use an improved CParser which can grok/skip more constructor flavours
        self.parser = CParserEnhanced(self.header_files, cache=self.cache_file)

    def parse_annotations(self):
        """
        Grok annotations like::

            // name=heading; expr=hdg * 20; unit=degrees
        """

        # compute list of source files (.h) with absolute paths
        source_files = [os.path.join(self.include_path, header_file) for header_file in self.header_files]

        # parse and compute annotations
        self.annotations = AnnotationParser(source_files, cache=self.cache_file + '.anno')

    def load(self):
        self.clib = CLibrary(self.library_file, self.parser, prefix='Lib_', lock_calls=False, convention='cdll', backend='ctypes')

    def struct_names(self):
        return self.parser.defs['structs'].keys()

    @classmethod
    def from_header(cls, include_path=None, header_files=None):
        cache_dir = user_cache_dir('lst', 'kotori')
        if not os.path.isdir(cache_dir): os.makedirs(cache_dir)
        library = LibraryAdapter(
            header_files, cls.compile(include_path, header_files),
            include_path=include_path, library_path=include_path,
            cache_path=cache_dir)
        return library

    @classmethod
    def compile(cls, include_path, header_files):

        # compute list of source files (.h) with absolute paths
        source_files = [os.path.join(include_path, header_file) for header_file in header_files]
        source_files = cls.augment_sources(source_files)
        source_files = ' '.join(source_files)

        # compute absolute path to library (.so) file
        library_file = header_files[0].rstrip('.h') + '.so'
        library_file = os.path.join(include_path, library_file)

        # assemble compiler command
        compiler = cls.find_compiler()
        if not compiler:
            raise ValueError('Could not find compiler')
        command = '{compiler} -std=c++11 -shared -fPIC -lm -o {library_file} {source_files}'.format(**locals())

        # run compile command
        logger.info(slm('Compiling: {}'.format(command)))
        retval = os.system(command)
        if retval == 0:
            logger.info(slm('Successfully compiled library "{library_file}" from "{source_files}"'.format(**locals())))
            return library_file
        else:
            msg = 'Failed compiling library "{library_file}" from "{source_files}"'.format(**locals())
            logger.error(slm(msg))
            raise ValueError(msg)

    @classmethod
    def find_compiler(cls):
        # TODO: make compiler configurable via ini file
        # TODO: investigate problem with /usr/bin/clang++: "clang: error: cannot specify -o when generating multiple output files"
        compilers = ['/opt/local/bin/g++-mp-4.7', '/usr/bin/g++']
        for compiler in compilers:
            if os.path.isfile(compiler):
                return compiler

    @classmethod
    def augment_sources(cls, source_files):
        """
        Augments source files and puts them under /tmp:
        - Automatically add ``#include "stdint.h"`` (required for types ``uint8_t``, etc.)
          and remove ``#include "mbed.h"`` (croaks on Intel)
        """
        source_files_augmented = []
        for source_file in source_files:

            # split source file into parts ...
            dirname = os.path.dirname(source_file)
            basename = os.path.basename(source_file)
            name, ext = os.path.splitext(basename)

            # and compute augmented file name
            name_intel = name
            name_intel += '_intel'
            name_intel += ext
            source_file_augmented = os.path.join(dirname, name_intel)

            # augment sourcecode
            source_payload = cls.augment_source(source_file)

            # write augmented file
            file(source_file_augmented, 'w').write(source_payload)
            source_files_augmented.append(source_file_augmented)

        return source_files_augmented

    @classmethod
    def augment_source(cls, source_file):
        """
        No matter if there's a ``#include "mbed.h"``, make sure it gets removed.
        Also make sure that there is a single ``#include "stdint.h"``.
        """
        payload = file(source_file).read()
        if re.search('^#include "stdint.h"', payload, re.MULTILINE):
            amendment = ''
        else:
            amendment = '#include "stdint.h"'

        if re.search('^#include "mbed.h"', payload, re.MULTILINE):
            payload = payload.replace('#include "mbed.h"', amendment)
        else:
            if amendment:
                payload = amendment + '\n' + payload
        return payload


class AnnotationParser(object):

    def __init__(self, source_files, cache=None):

        self.source_files = source_files

        # keep all parsed information here
        self.info = {'structs': {}}

        # parse all annotations
        self.parse()

    def parse(self):
        """
        Parse rule annotations like::

                int16_t  hdg                ;//6        /* @rule: name=heading; expr=hdg * 20; unit=degrees */

        """

        # e.g. "typedef struct struct_position         // added 06.03.2014 C.L."
        struct_head_pattern  = re.compile('^.*struct\s+([\w]+).*?$')

        # e.g. "/* @rule: name=heading; expr=hdg * 20; unit=degrees */"
        rule_extract_pattern = re.compile('^.*name=(?P<name>\w+?);.*expr=(?P<expression>.+?);.*unit=(?P<unit>\w+).*$')

        # e.g. "   int16_t  hdg                ;"
        original_fieldname_pattern = re.compile('^.*?(?P<type>[\w]+)\s+(?P<name>[\w]+).*;.*$')

        for source_file in self.source_files:

            struct_name = None
            with file(source_file) as f:
                for line in f.readlines():
                    line = line.strip()

                    # check for beginning of struct
                    m = struct_head_pattern.match(line)
                    if m:
                        struct_name = m.group(1)

                    # check for transformation rule annotation
                    if '@rule:' in line:
                        m = rule_extract_pattern.match(line)
                        if m:
                            rule = m.groupdict()
                            if 'expression' in rule:
                                rule['expression_sympy'] = sympify(rule['expression'])

                            # parse original field name to map against
                            m = original_fieldname_pattern.match(line)
                            if m:
                                vanilla_field = m.groupdict()
                                field_name = vanilla_field['name']
                                self.info['structs'].setdefault(struct_name, {})
                                self.info['structs'][struct_name][field_name] = rule

        #pprint(self.info)

    def get_struct_rules(self, struct_name):
        return self.info['structs'].get(struct_name, {})

    def get_struct_rule(self, struct_name, field_name):
        struct_rules = self.get_struct_rules(struct_name)
        return struct_rules.get(field_name)


class StructAdapter(object):

    def __init__(self, name, library):
        self.name = name
        self.library = library
        self.parser = self.library.parser
        self.clib = self.library.clib
        self.annotations = self.library.annotations

    def ast(self):
        return self.parser.defs['structs'][self.name]

    def obj(self):
        return self.clib('structs', self.name)

    def create(self):
        """
        Create and initialize ctypes struct.
        """
        # wrapper class from pyclibrary.backend_ctypes
        obj_wrapper = self.obj()

        # <ctypes struct 'struct_foo'>
        obj = obj_wrapper()

        # initialize struct content
        self.initialize(obj)

        return obj

    def initialize(self, obj):
        """
        Initialize struct instance with initializer data
        """
        data = self._get_initializer_data()
        if data:
            bytes = ''.join(map(chr, data))
            obj._load_(bytes)

    def _get_initializer_data(self):
        """
        Nasty hack to get proper initializer data from CParser results.

        This works by traversing the CParser result nodes from the variable ``position``
        back to the struct ``struct_position`` to correlate struct with initializer data.

        Example struct declaration::

            struct struct_position
            {
                uint8_t  length             ;//1
                uint8_t  ID                 ;//2
                // ...
            } position = {9,1};

        Obviously, this will only work for singletons, where each struct used is instantiated only once.
        You have been warned.
        """
        for key, value in self.parser.defs['variables'].iteritems():
            data, type = value
            type_spec = 'struct ' + self.name
            if type_spec == type.type_spec:
                return data

        # don't log, this would be triggered erroneously when using equal-style-initializers
        #logger.warn('Could not find initialization data for struct "{}"'.format(self.name))

    def __repr__(self):
        return "<StructAdapter '{name}' object at {id}>".format(name=self.name, id=hex(id(self)))

    def print_schema(self):
        width = 84

        def print_header(text):
            print '-' * width
            print text.center(width)
            print '-' * width
            print

        print_header('struct "{}"'.format(self.name))

        print 'Header information'
        print
        print tabulate(self.ast_members_repr(), headers=['name', 'type', 'default', 'bitfield'])
        print; print

        print 'Library information'
        print
        print tabulate(self.lib_fields_repr(), headers=['name', 'type', 'symbol', 'field', 'bitfield'])
        print; print

        print 'Transformation rules'
        print
        print tabulate(self.annotations_repr(), headers=['name-real', 'name-human', 'expression', 'expression-sympy', 'unit'])
        print; print

        print 'Representations'
        print
        print tabulate(self.lib_binary_repr(), headers=['kind', 'representation'])
        print; print

    def ast_members_repr(self):

        def ast_type_repr(t):
            type_qual_str = ('' if not any(t.type_quals) else
                             ', type_quals='+repr(t.type_quals))
            return (', '.join(map(str, t)) + type_qual_str)

        members = []
        for member in self.ast().members:
            member_length = len(member)
            bitfield = None
            if member_length == 3:
                name, type, default = member
            elif member_length == 4:
                name, type, default, bitfield = member
            else:
                raise ValueError("Don't know how to handle ast members with length={}".format(member_length))

            # use name of type object
            type = ast_type_repr(type)

            entry = (name, type, default, bitfield)
            members.append(entry)

        return members

    def lib_fields_repr(self):
        fields = []
        o = self.obj()
        for field_item in o._fields_:

            field_length = len(field_item)
            bitfield = None
            if field_length == 2:
                name, ctype = field_item
            elif field_length == 3:
                name, ctype, bitfield = field_item
            else:
                raise ValueError("Don't know how to handle lib fields with length={}".format(field_length))

            # use name of ctypes object
            ctype_name = ctype.__name__
            symbol = ctype._type_

            # get ctypes Field object
            field = getattr(o, name)

            entry = (name, ctype_name, symbol, str(field), bitfield)
            fields.append(entry)

        return fields

    def lib_binary_repr(self):
        payload = self.create()._dump_()
        return self.binary_reprs(payload)

    def annotations_repr(self):
        entries = []
        annos = self.annotations.get_struct_rules(self.name)
        if annos:
            for fieldname, rule in annos.iteritems():
                entry = (fieldname, rule['name'], rule['expression'], rule['expression_sympy'], rule['unit'])
                entries.append(entry)
        return entries

    @staticmethod
    def binary_reprs(payload):
        reprs = [
            ('hex',     '0x' + hexlify(payload)),
            ('decimal', map(ord, payload)),
            ('bytes',   repr(payload)),
        ]
        return reprs


class StructRegistry(object):

    def __init__(self, library):
        self.library = library

        # dictionary of structs, by name
        self.structs = {}

        self.build()

    def build(self):
        for name in self.library.struct_names():
            logger.debug('Rolling in struct "{}"'.format(name))
            self.register_adapter(name)

    def register_adapter(self, name):
        # store adapter object by name
        adapter = StructAdapter(name, self.library)
        self.structs[name] = adapter
        return adapter

    def get(self, name):
        return self.structs[name]

    def get_by_id(self, struct_id):
        try:
            return self.structs_by_id[struct_id]
        except KeyError:
            msg = 'Struct with id {} ({}) not registered.'.format(struct_id, hex(struct_id))
            logger.error(slm(msg))
            raise KeyError(msg)

    def create(self, name, **attributes):
        return self.get(name).create(**attributes)

    @classmethod
    def to_dict(cls, struct):
        # TODO: maybe refactor to struct._dict_
        fieldnames = [field[0] for field in struct._fields_]
        d = OrderedDict()
        for fieldname in fieldnames:
            d[fieldname] = getattr(struct, fieldname)
        return d

    @classmethod
    def pprint(cls, struct, data=None, format='pprint'):
        # TODO: maybe refactor to struct._pprint_
        name = struct._name_()
        payload = struct._dump_()
        payload_hex = hexlify(payload)
        if data:
            payload_data = list(data.items())
        else:
            payload_data = list(cls.to_dict(struct).items())

        if format == 'pprint':
            print 'name:', name
            print 'hex: ', payload_hex
            pprint(payload_data, indent=4, width=42)

        elif format == 'tabulate-plain':
            separator = ('----', '')
            output = [
                separator,
                ('name', name),
                separator,
            ]
            output += StructAdapter.binary_reprs(payload)
            output += [separator]
            output += payload_data
            print tabulate(output, tablefmt='plain')

            #print tabulate(list(meta.items()), tablefmt='plain')
            #print tabulate(payload_data, missingval='n/a', tablefmt='simple')

        else:
            raise ValueError('Unknown format "{}" for pretty printer'.format(format))



class StructRegistryByID(StructRegistry):

    # TODO: Make generic like ``sr = StructRegistryByID(library, indexes=['ID'])``

    def __init__(self, library, indexes=None):
        # dictionary of structs, by id
        self.structs_by_id = {}
        StructRegistry.__init__(self, library)

    def register_adapter(self, name):

        adapter = StructRegistry.register_adapter(self, name)

        # store adapter object by id
        struct_id = adapter.create().ID
        if self.structs_by_id.has_key(struct_id):
            o_a = self.structs_by_id[struct_id]
            name_owner = o_a.name
            logger.warn('Struct "{}" has ID "{}", but this ID is already mapped to struct "{}", '\
                           'please check if struct provides reasonable default values for attribute "ID"'.format(name, struct_id, name_owner))
        else:
            logger.debug('Struct "{}" mapped to ID "{}"'.format(name, struct_id))
            self.structs_by_id[struct_id] = adapter

    def get_metadata(self):
        metadata = []
        for index in sorted(self.structs_by_id.keys()):
            struct = self.structs_by_id[index]
            members = struct.ast_members_repr()
            fieldnames = [member[0] for member in members]
            entry = OrderedDict()
            entry['id'] = index
            entry['name'] = struct.name
            entry['# structs'] = len(fieldnames)
            metadata.append(entry)
        return metadata



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@getkotori.org>
import json
import delegator

def publish(topic, message):
    command = 'mosquitto_pub -h "{host}" -t "{topic}" -m \'{message}\''.format(
        host='localhost', topic=topic, message=message)

    print 'Running command: {}'.format(command)
    delegator.run(command)

def all_realms():
    #realms = ['mqttkit-1', 'hiveeyes', 'luftdaten', 'weewx']
    realms = ['hiveeyes']
    measurements = {"temperature": 42.84, "weight": 84.85}
    for realm in realms:
        topic = '{}/testdrive/area-42/node-1/data.json'.format(realm)
        publish(topic, json.dumps(measurements))

def hiveeyes_testcases():

    # Publish en bloc using German field names
    #data = json.dumps({u'Gewicht 1': 5.0, u'Gewicht Total': 42.42, u'Außentemperatur': 32.32})
    #publish('hiveeyes/testdrive/area-42/node-1/data.json', data)

    # Don't confuse field name classifier
    #data = json.dumps({u'Weight Total StdDev': 42.42, u'Weight Total': 42.42, u'Temperature 8 Outside': 32.32})
    #publish('hiveeyes/testdrive/area-42/node-2/data.json', data)

    # Publish discrete values I (from deprecated firmware)
    """
    hiveeyes/kh/cfb/hive1/measure/weight   55.435
    hiveeyes/kh/cfb/hive1/measure/broodtemperature  35.2
    hiveeyes/kh/cfb/hive1/measure/entrytemperature   6.6
    hiveeyes/kh/cfb/hive1/measure/airtemperature  21.7
    hiveeyes/kh/cfb/hive1/measure/airhumidity  58.5
    hiveeyes/kh/cfb/hive1/measure/airtemperature_outside (null)
    hiveeyes/kh/cfb/hive1/measure/airhumidity_outside (null)
    """
    #publish('hiveeyes/testdrive/area-42/node-3/measure/weight', 55.435)
    #publish('hiveeyes/testdrive/area-42/node-3/measure/airtemperature', 21.7)

    # Publish discrete values II (via LoRaWAN/TTN)
    """
    hiveeyes/thias/thias-hive2/up/data/relative_humidity_2 51.5
    hiveeyes/thias/thias-hive2/up/data/analog_in_3 3.5
    hiveeyes/thias/thias-hive2/up/data/analog_in_1 -13.54
    hiveeyes/thias/thias-hive2/up/data/temperature_1 21.1
    hiveeyes/thias/thias-hive2/up/data/relative_humidity_1 48.5
    hiveeyes/thias/thias-hive2/up/data/temperature_2 20.8
    """
    #publish('hiveeyes/testdrive/area-42/node-4/data/analog_in_3', 3.5)
    #publish('hiveeyes/testdrive/area-42/node-4/data/temperature_1', 21.1)


    # Special characters in topic name
    """
    data = json.dumps({"Weight Total StdDev": -0.017, "Temperature 3 Inside": "  7.00", "Humidity Outside": "99.90",
                       "Temperature 8 Outside": "3.60", "Temperature 1 Inside": "  6.69", "Temperature 2 Inside": "  3.56",
                       "Yield Sum": -0.821, "Weight Total": 25.599, "Weight 1": 4.827, "Weight 2": 8.45, "Weight 3": 5.202,
                       "Weight 4": 7.12})
    publish('hiveeyes/testdrive/area@42/node-5/data.json', data)
    """


def run():

    # Publish something to all configured realms
    #all_realms()

    # Runbook for some acquisition feeds arriving for the "hiveeyes" realm
    #hiveeyes_testcases()

    # weewx problem
    # 2018-04-09T07:20:15+0200 [kotori.daq.services.mig            ] WARN    : Unknown message type on topic "pressure_mbar" with payload "937.860347662"
    # weewx/097287c4-6fb0-4aeb-a095-00d65ecb15f7/Leoni/VantagePro2/pressure_mbar 937.860347662
    publish('weewx/testdrive/area-42/node-1/pressure_mbar', 937.860347662)



# -*- coding: utf-8 -*-
# (c) 2019 Andreas Motl <andreas@getkotori.org>
import types
from collections import OrderedDict


class TasmotaDecoder:

    def __init__(self, topology):
        self.topology = topology

    def decode_message(self, message):
        # Fixme: Currently ignores the "Time" field.
        if 'slot' in self.topology and self.topology.slot.endswith('SENSOR'):
            message = self.decode_sensor_message(message)
        if 'slot' in self.topology and self.topology.slot.endswith('STATE'):
            message = self.decode_state_message(message)
        return message

    def decode_sensor_message(self, message):
        """
        {
          "Time": "2019-06-02T22:13:07",
          "SonoffSC": {
            "Temperature": 25,
            "Humidity": 15,
            "Light": 20,
            "Noise": 10,
            "AirQuality": 90
          },
          "TempUnit": "C"
        }

        {
          "Time": "2017-02-16T10:13:52",
          "DS18B20": {
            "Temperature": 20.6
          }
        }

        """
        data = OrderedDict()
        for key, value in message.items():
            if isinstance(value, types.DictionaryType):
                data.update(value)
        return data

    def decode_state_message(self, message):
        """
        {
          "Time": "2019-06-02T22:13:07",
          "Uptime": "1T18:10:35",
          "Vcc": 3.182,
          "SleepMode": "Dynamic",
          "Sleep": 50,
          "LoadAvg": 19,
          "Wifi": {
            "AP": 1,
            "SSId": "{redacted}",
            "BSSId": "A0:F3:C1:{redacted}",
            "Channel": 1,
            "RSSI": 100,
            "LinkCount": 1,
            "Downtime": "0T00:00:07"
          }
        }
        """
        data = OrderedDict()
        data['Device.Vcc'] = message.get('Vcc')
        data['Device.Sleep'] = message.get('Sleep')
        data['Device.LoadAvg'] = message.get('LoadAvg')
        data['Device.Wifi.Channel'] = message.get('Wifi', {}).get('Channel')
        data['Device.Wifi.RSSI'] = message.get('Wifi', {}).get('RSSI')
        data['Device.Wifi.LinkCount'] = message.get('Wifi', {}).get('LinkCount')
        return data



# -*- coding: utf-8 -*-
# (c) 2014-2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import sys
from twisted.internet.defer import inlineCallbacks
from twisted.python import log
from twisted.internet import reactor
from autobahn.twisted.wamp import ApplicationRunner, ApplicationSession

class KotoriClient(ApplicationSession):
    """
    Simple message publishing with Autobahn WebSockets.

    derived from:
    - https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/basic/pubsub/basic/frontend.py
    - https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/pubsub/simple/example2/client.py
    """

    @inlineCallbacks
    def onJoin(self, details):
        print("session attached")

        yield self.publish("de.elmyra.kotori.broadcast.node-activity", {'node_id': 'Hello!', 'state': True}, excludeMe=False)
        yield self.publish("de.elmyra.kotori.broadcast.operator-presence", True, excludeMe=False)
        #yield self.publish("de.elmyra.kotori.broadcast.operator-presence", False, excludeMe=False)

        # send telemetry data
        yield self.publish("de.elmyra.kotori.telemetry.data", 'WAMP hello', excludeMe=False)
        yield self.publish("de.elmyra.kotori.telemetry.data", {'attribute1': 'value1', 'attribute2': 'value2'}, excludeMe=False)

        reactor.callLater(0.1, reactor.stop)

    def onDisconnect(self):
        print("disconnected")
        reactor.stop()

def run_wamp_client():
    log.startLogging(sys.stdout)
    runner = ApplicationRunner("ws://localhost:9000/ws", "kotori-realm")
    #runner = ApplicationRunner("ws://master.example.com:9000/ws", "kotori-realm")
    runner.run(KotoriClient, start_reactor=False)
    reactor.run()


if __name__ == '__main__':
    run_wamp_client()



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import sys
from twisted.logger import Logger, LogLevel
from twisted.internet import reactor
from autobahn.twisted.wamp import ApplicationRunner, ApplicationSession
from kotori.util.logger import startLogging

log = Logger()

class WampSession(ApplicationSession):
    """
    Simple message publishing with Autobahn WebSockets.

    derived from:
    - https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/basic/pubsub/basic/frontend.py
    - https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/pubsub/simple/example2/client.py
    - https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/session/fromoutside/client.py
    """

    component = None

    def onJoin(self, details):
        log.info("WAMP session joined: {}".format(details))
        try:
            if self.component:
                self.component(bus=self, config=self.config.extra)
        except Exception:
            log.failure('Could not create downstream component after joining WAMP bus')

    def onDisconnect(self):
        log.info("WAMP session disconnected")


class WampApplication(object):

    def __init__(self, url, realm=u'kotori', session_class=WampSession, config=None):
        """
        url: ws://master.example.com:9000/ws
        """
        self.url = url
        self.realm = realm
        self.session_class = session_class
        self.config = config

    def make(self):

        # connect to crossbar router/broker
        self.runner = ApplicationRunner(self.url, self.realm, extra=dict(self.config))

        # run application session
        self.deferred = self.runner.run(self.session_class, start_reactor=False)

        def croak(ex, *args):
            log.error('Problem in {name}, please check if "crossbar" WAMP broker is running. args={args}'.format(
                name=self.__class__.__name__, args=args))
            log.error("{ex}, args={args!s}", ex=ex.getTraceback(), args=args)
            reactor.stop()
            raise ex

        self.deferred.addErrback(croak)


if __name__ == '__main__':
    startLogging(sys.stdout, level=LogLevel.levelWithName('debug'))
    app = WampApplication(url=u'ws://localhost:9000/ws')
    app.make()
    reactor.run()



# -*- coding: utf-8 -*-
# (c) 2016-2018 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
# https://pypi.python.org/pypi/paho-mqtt/
from __future__ import absolute_import
import os
import paho.mqtt.client as mqtt
from twisted.logger import Logger
from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from twisted.application.service import Service
from kotori.daq.intercom.mqtt.base import BaseMqttAdapter

log = Logger()

class PahoMqttAdapter(BaseMqttAdapter, Service):

    # If connection fails at first, retry connecting each X seconds
    retry_interval = 5

    def connect(self):
        """
        Connect to MQTT broker.
        """
        # TODO: Check if we can do asynchronous connection establishment.
        #       Currently, this is done synchronously which could harm
        #       other subsystems in timeout or otherwise blocking situations.

        # Make MQTT client identifier even more unique by adding process id
        pid = os.getpid()
        client_id = '{}:{}'.format(self.name, str(pid))

        # Connection establishment
        self.client = mqtt.Client(client_id=client_id, clean_session=True)

        # Optionally authenticate connection
        if self.broker_username:
            self.client.username_pw_set(self.broker_username, self.broker_password)

        # Set event handlers
        self.client.on_connect = lambda *args: reactor.callFromThread(self.on_connect, *args)
        self.client.on_message = lambda *args: reactor.callFromThread(self.on_message, *args)
        self.client.on_log     = lambda *args: reactor.callFromThread(self.on_log, *args)

        # Connect with retry
        self.connect_loop = LoopingCall(self.connect_with_retry)
        self.connect_loop.start(self.retry_interval, now=True)

    def connect_with_retry(self):
        try:
            self.client.connect(self.broker_host, port=self.broker_port, keepalive=60)
            self.connect_loop.stop()
        except:
            log.failure(u'Error connecting to MQTT broker but retrying each {retry_interval} seconds',
                retry_interval=self.retry_interval)
            return

        """
        This is part of the threaded client interface. Call this once to
        start a new thread to process network traffic. This provides an
        alternative to repeatedly calling loop() yourself.
        """
        self.client.loop_start()
        reactor.addSystemEventTrigger('before', 'shutdown', self.client.loop_stop, True)

    # The callback for when the client receives a CONNACK response from the server.
    def on_connect(self, client, userdata, flags, rc):
        """
        on_connect(client, userdata, flags, rc): called when the broker responds to our connection
          request.
          flags is a dict that contains response flags from the broker:
            flags['session present'] - this flag is useful for clients that are
                using clean session set to 0 only. If a client with clean
                session=0, that reconnects to a broker that it has previously
                connected to, this flag indicates whether the broker still has the
                session information for the client. If 1, the session still exists.
          The value of rc determines success or not:
            0: Connection successful
            1: Connection refused - incorrect protocol version
            2: Connection refused - invalid client identifier
            3: Connection refused - server unavailable
            4: Connection refused - bad username or password
            5: Connection refused - not authorised
            6-255: Currently unused.
        """
        log.debug("Connected to MQTT. userdata={userdata}, flags={flags}, rc={rc}",
            userdata=userdata, flags=flags, rc=rc)

        # Subscribing in on_connect() means that if we lose the connection and
        # reconnect then subscriptions will be renewed.
        #client.subscribe("$SYS/#")
        self.subscribe()

    # The callback for when a PUBLISH message is received from the server.
    def on_message(self, client, userdata, message):
        """
        on_message(client, userdata, message): called when a message has been received on a
          topic that the client subscribes to. The message variable is a
          MQTTMessage that describes all of the message parameters.
        """

        # TODO: Do something with "client" object (paho.mqtt.client.Client)

        topic = message.topic
        payload = message.payload

        # Make metadata dictionary to be passed as kwargs later

        # Mungle topic and payload out of metadata
        metadata = message.__dict__.copy()
        del metadata['topic']
        del metadata['payload']

        # Mungle userdata into message
        metadata['userdata'] = userdata

        if not topic.endswith('error.json'):
            log.debug('on_message: name={name}, topic={topic}, payload={payload}, kwargs={kwargs}', name=self.name, topic=topic, payload=payload, kwargs=metadata)

        return self.callback(topic=topic, payload=payload, **metadata)

    def publish(self, topic, payload):
        log.debug(u'Publishing to topic={topic}, payload={payload}', topic=topic, payload=payload)
        return self.client.publish(topic, payload)

    def subscribe(self, *args):
        #d = self.protocol.subscribe("foo/bar/baz", 0)
        log.info(u"Subscribing to topics {subscriptions}. client={client}", subscriptions=self.subscriptions, client=self.client)
        for topic in self.subscriptions:
            log.info(u"Subscribing to topic '{topic}'", topic=topic)
            # Topic name **must not** be unicode, so casting to string
            e = self.client.subscribe(str(topic), qos=0)

    def on_log(self, client, userdata, level, buf):
        """
        on_log(client, userdata, level, buf): called when the client has log information. Define
          to allow debugging. The level variable gives the severity of the message
          and will be one of MQTT_LOG_INFO, MQTT_LOG_NOTICE, MQTT_LOG_WARNING,
          MQTT_LOG_ERR, and MQTT_LOG_DEBUG. The message itself is in buf.
        """
        log.debug(u'{message}. level={level_mqtt}, userdata={userdata}', message=buf, level_mqtt=level, userdata=userdata)



# -*- coding: utf-8 -*-
# (c) 2015-2018 Andreas Motl <andreas@getkotori.org>
from copy import deepcopy

from bunch import Bunch
from twisted.application.service import MultiService

class MultiServiceMixin(object):

    def __init__(self, name=None):
        MultiService.__init__(self)
        self.name = name

    def startService(self):
        try:
            self.setupService()
        except AttributeError:
            pass
        MultiService.startService(self)

    def registerService(self, service):
        service.setServiceParent(self)

    def setupChannel(self, channel=None, name=None):
        """
        Make channel object from application settings configuration object
        """

        assert channel is not None, 'Channel object must not be None'

        channel = deepcopy(channel)
        if name:
            channel.update(name=name)
        self.channel = Bunch(**channel)

    @property
    def logname(self):
        return u'{class_name}({name})'.format(class_name=self.__class__.__name__, name=self.name)

    def log(self, level, prefix):
        level('{prefix} {class_name}. name={name}, channel={channel}',
            prefix=prefix, class_name=self.__class__.__name__, name=self.name, channel=dict(self.channel))

class RootService(MultiService, MultiServiceMixin):

    def __init__(self, settings=None):
        MultiService.__init__(self)
        self.settings = settings

    def makeService(self, service_class, *args, **kwargs):
        service = service_class(*args, **kwargs)
        self.registerService(service)
        return service



# -*- coding: utf-8 -*-
# (c) 2015-2018 The Hiveeyes Developers <hello@hiveeyes.org>
import re
import json

import attr
from bunch import Bunch
from collections import OrderedDict

from grafana_api_client import GrafanaPreconditionFailedError, GrafanaClientError
from pkg_resources import resource_filename
from jinja2 import Template
from twisted.logger import Logger
from kotori.daq.services import RootService
from kotori.daq.services.mig import MqttInfluxGrafanaService
from kotori.daq.intercom.strategies import WanBusStrategy
from kotori.daq.graphing.grafana.manager import GrafanaManager
from kotori.util.common import SmartBunch

log = Logger()

class HiveeyesGenericGrafanaManager(GrafanaManager):

    knowledge = [
        {'name': 'temperature', 'prefixes': ['temp', 'Temp'],   'suffixes': ['temp', 'temperature'],    'format': 'celsius'},
        {'name': 'humidity',    'prefixes': ['hum'],            'suffixes': ['hum', 'humidity'],        'format': 'humidity'},
        {'name': 'weight',      'prefixes': ['wght', 'weight', 'Gewicht', 'Weight'],    'label':  'kg'},
        {'name': 'volume',      'prefixes': ['volume'],                                 'label':  'dB', 'scale': 10},
    ]

    def get_dashboard_identity(self, storage_location, topology=None):

        # Compute effective topology information
        topology = topology or {}
        realm = topology.get('realm', 'default')
        network = topology.get('network', storage_location.database)

        # Derive dashboard uid and name from topology information
        identity = SmartBunch(
            uid=u'{realm}-{network}-instant'.format(realm=realm, network=network),
            name=u'{realm}-{network}'.format(realm=realm, network=network),
            title=u'{realm}-{network}'.format(realm=realm, network=network),
            # TODO: Use real title after fully upgrading to new Grafana API (i.e. don't use get-by-slug anymore!)
            #title=u'Hiveeyes Rohdaten im Netzwerk ' + network,
        )
        #print identity.prettify()

        return identity

    def get_rule(self, fieldnames):
        """
        Filter knowledgebase by prefix.
        Obtains a list of fieldnames / fieldname prefixes and
        also matches against a list of fieldname prefixes,
        hence the nested loops.
        """

        if not fieldnames:
            return {}

        for rule in self.knowledge:
            for prefix in rule['prefixes']:
                for fieldname in fieldnames:
                    if fieldname.startswith(prefix):
                        return rule

        return {}

    def get_panel_options(self, data, fieldnames):
        return self.get_rule(fieldnames)

    def panel_title_prefix(self, fieldnames):
        return self.get_rule(fieldnames).get('name')

    def get_distinct_panel_field_prefixes(self, data):
        """
        Compute list of fieldname prefixes in order of knowledgebase
        for providing stable input to the panel generator.
        """

        prefixes = OrderedDict()
        fields_given = data.keys()
        fields_used = []
        for rule in self.knowledge:
            for prefix in rule['prefixes']:
                for fieldname in fields_given:
                    if fieldname.startswith(prefix):
                        key = '-'.join(rule['prefixes'])
                        prefixes[key] = rule['prefixes']
                        fields_used.append(fieldname)

            if 'suffixes' in rule:
                for suffix in rule['suffixes']:
                    for fieldname in fields_given:
                        if fieldname.endswith(suffix):
                            key = '-'.join(rule['prefixes'])
                            prefixes[key] = rule['suffixes']
                            fields_used.append(fieldname)

        # Add unused fields
        fields_unused = list(set(fields_given) - set(fields_used))
        if fields_unused:
            prefixes['misc'] = fields_unused

        return prefixes.values()

    def panel_generator(self, storage_location, data, topology):
        """
        Generate Hiveeyes-specific Grafana panels, which is:
        Group multiple temperature or humidity sensors into a single panel (aka. one panel per sensor family),
        even when transmitted in BERadio-specific shortcut notation like 'temp1', 'temp2', etc.
        """

        panels = []

        for prefixes in self.get_distinct_panel_field_prefixes(data):
            panel = self.get_panel_data(storage_location, topology, data, fieldname_prefixes=prefixes)
            panels.append(panel)

        return panels


class HiveeyesBeehiveGrafanaManager(GrafanaManager):

    def __init__(self, *args, **kwargs):
        GrafanaManager.__init__(self, *args, **kwargs)
        self.tpl_dashboard_weef = self.get_template('grafana-dashboard-weef-de.json')

    def get_template(self, filename):
        return Template(file(resource_filename('kotori.vendor.hiveeyes', filename)).read().decode('utf-8'))

    def get_dashboard_identity(self, storage_location, topology=None):

        # Compute effective topology information
        topology = topology or {}
        realm = topology.get('realm', 'default')
        network = topology.get('network', storage_location.database)

        # Derive dashboard uid and name from topology information
        nodename = u'{gateway} / {node}'.format(gateway=topology.gateway, node=topology.node)
        identity = SmartBunch(
            #uid=u'{realm}-{network}-instant'.format(realm=realm, network=network),
            name=self.strategy.topology_to_label(topology),
            title=self.strategy.topology_to_label(topology),
            # TODO: Use real title after fully upgrading to new Grafana API (i.e. don't use get-by-slug anymore!)
            # title=u'Hiveeyes Umweltcockpit für Meßknoten {nodename} im Netzwerk {network}'.format(nodename=nodename, network=network),
            # title=u'Hiveeyes Ertragscockpit für Meßknoten {nodename} im Netzwerk {network}'.format(nodename=nodename, network=network),
        )
        #print identity.prettify()

        return identity

    def provision(self, storage_location, message, topology):

        topology = topology or {}

        # The identity information of this provisioning process
        dashboard_identity = self.get_dashboard_identity(storage_location, topology)
        signature = (storage_location.database, storage_location.measurement)
        whoami = 'dashboard "{dashboard_name}" for database "{database}" and measurement "{measurement}"'.format(
            dashboard_name=dashboard_identity.name, database=storage_location.database, measurement=storage_location.measurement)

        dashboard_name = dashboard_identity.name

        # Skip dashboard creation if it already has been created while Kotori is running
        # TODO: Improve locking to prevent race conditions.
        if self.keycache.exists(*signature):
            log.debug('Data signature not changed, skip update of {whoami}', whoami=whoami)
            return

        log.info('Provisioning Grafana {whoami}', whoami=whoami)

        # Create a Grafana datasource object for designated database
        datasource_name = self.create_datasource(storage_location)

        # Try to classify designated field names from field names in message payload
        fieldnames = sorted(message.keys())
        computed_fields = BeekeeperFields(fieldnames=fieldnames).classify()

        # Create appropriate Grafana dashboard
        data_dashboard = {
            #'uid': dashboard_name + '-environment',
            'title': dashboard_identity.title,
            'datasource': datasource_name,
            'measurement_sensors': storage_location.measurement,
            'measurement_events': storage_location.measurement_events,
            'field': computed_fields,
        }

        # Debugging
        #pprint(data_dashboard)

        # Apply dashboard data to JSON template
        dashboard_json_weef = self.tpl_dashboard_weef.render(data_dashboard)

        # Get or create Grafana folder for stuffing all instant dashboards into
        folder = self.grafana_api.ensure_instant_folder()
        folder_id = folder and folder.get('id') or None

        # Create Grafana dashboards
        for dashboard_json in [dashboard_json_weef]:

            try:
                log.info('Creating/updating dashboard "{}"'.format(dashboard_name))
                response = self.grafana_api.grafana_client.dashboards.db.create(
                    folderId=folder_id, dashboard=json.loads(dashboard_json), overwrite=True)
                log.info(u'Grafana response: {response}', response=json.dumps(response))

            except GrafanaPreconditionFailedError as ex:
                if 'name-exists' in ex.message or 'A dashboard with the same name already exists' in ex.message:
                    log.warn(ex.message)
                else:
                    log.error('GrafanaPreconditionFailedError: {ex}', ex=ex.message)

            except GrafanaClientError as ex:
                log.error('GrafanaClientError: {ex}', ex=ex.message)

        # Remember dashboard/panel creation for this kind of data inflow
        self.keycache.set(storage_location.database, storage_location.measurement)


@attr.s
class BeekeeperFields(object):

    fieldnames = attr.ib()

    def classify(self):

        log.debug(u'Classifying beekeeper fields {fieldnames}', fieldnames=self.fieldnames)

        # TODO: Can we account for multiple occurrences of "weightX" fields for mapping more than one scale?

        weight_synonyms = u'(weight|wght|gewicht)'
        temperature_synonyms = u'(temperature|temp|temperatur)'
        outside_synonyms = u'(outside|out|air|außen|aussen)'

        candidates = {
            'weight_total': [
                self.from_words(weight_synonyms, 'total', exclude=['stddev']),
                self.from_words(weight_synonyms, exclude=['stddev']),
            ],
            'temperature_outside': [
                self.from_words(temperature_synonyms, outside_synonyms),
                self.from_words(temperature_synonyms),
            ],
            'temperature_inside': [
                self.from_words(temperature_synonyms, 'inside'),
                self.from_words(temperature_synonyms),
            ],
        }
        #pprint(candidates)

        results = SmartBunch()
        for name, patterns in candidates.items():
            fieldname = self.find_match(patterns)
            if fieldname is not None:
                results[name] = fieldname

        log.info(u'Classified beekeeper fields "{fields}" from "{fieldnames}"', fields=results.dump(), fieldnames=self.fieldnames)

        return results

    @staticmethod
    def from_words(*words, **kwargs):
        """
        Positive and Negative Lookahead
        https://www.regular-expressions.info/lookaround.html

        Positive lookahead
        Synopsis: q(?=u)
        Example:  (?=^.*?foo.*$)(?=^.*?bar.*$)(?=^.*?green.*$)^.*$

        Negative lookahead
        Synopsis: q(?!u)
        """
        patterns = []
        for word in words:
            patterns.append(u'(?=^.*?{word}.*$)'.format(word=word))
        if 'exclude' in kwargs:
            for word in kwargs['exclude']:
                patterns.append(u'(?!^.*?{word}.*$)'.format(word=word))
        pattern = u''.join(patterns) + u'^.*$'
        return pattern

    def find_match(self, patterns):
        # TODO: Should optimize this - don't compile regex patterns on each invocation!
        for pattern in patterns:
            matcher = re.compile(pattern, re.IGNORECASE)
            for fieldname in self.fieldnames:
                if matcher.search(fieldname):
                    return fieldname



def hiveeyes_boot(settings, debug=False):

    # Service container root
    rootService = RootService(settings=settings)

    # Channel realm
    channel = Bunch(**settings.hiveeyes)

    # Main application service
    service = MqttInfluxGrafanaService(
        channel  = channel,
        graphing = [
            HiveeyesGenericGrafanaManager(settings=settings, channel=channel),
            HiveeyesBeehiveGrafanaManager(settings=settings, channel=channel),
        ],
        strategy = WanBusStrategy()
    )

    rootService.registerService(service)
    rootService.startService()



# -*- coding: utf-8 -*-
# (c) 2016-2017 Andreas Motl <andreas@getkotori.org>
from pyramid.settings import asbool
from twisted.internet import threads
from twisted.web import http, server
from twisted.logger import Logger, LogLevel
from twisted.application.service import MultiService
from kotori.util.configuration import read_list
from kotori.daq.services import MultiServiceMixin
from kotori.daq.intercom.mqtt import MqttAdapter
from kotori.io.protocol.http import HttpDataFrameResponse
from kotori.io.protocol.util import handleFailure
from kotori.util.errors import last_error_and_traceback

log = Logger()

try:
    from kotori.io.export.influx import DataFrameQuery
except ImportError:
    log.failure('InfluxDB export not available, please install "pandas".', level=LogLevel.warn)

class ForwarderTargetService(MultiServiceMixin, MultiService):
    """
    Container service for target services.

    As of June 2016, there are currently two target
    services for emitting data, MQTT and InfluxDB.
    """

    def __init__(self, address=None, **kwargs):
        MultiServiceMixin.__init__(self, **kwargs)

        self.address = address
        self.scheme  = self.address.uri.scheme

        self.downstream = None

    def setupService(self):
        """
        Configure effective data target by registering an appropriate
        downstream service object for handling the target address scheme.
        """

        log.info(u'Starting {name} for serving address {address}', name=self.logname, address=self.address)

        self.settings = self.parent.settings

        if self.scheme == 'mqtt':

            # Register MqttAdapter service as downstream subsystem service object
            self.downstream = MqttAdapter(
                name          = self.name + '-downstream',
                broker_host   = self.settings.mqtt.host,
                broker_port   = int(self.settings.mqtt.port),
                broker_username = self.settings.mqtt.username,
                broker_password = self.settings.mqtt.password)

        elif self.scheme == 'influxdb':
            # InfluxDB has no subsystem service, it's just an adapter
            pass

        else:
            raise KeyError('No target/downstream dispatcher for scheme {scheme}'.format(scheme=self.scheme))

        # Register service component with its container
        if self.downstream:
            self.registerService(self.downstream)

    def emit(self, uri, bucket):
        """
        Adapt, serialize and emit data bucket to target service.
        """

        log.debug('Emitting to target scheme {scheme}', scheme=self.scheme)

        if self.scheme == 'mqtt':

            # Publish JSON payload to MQTT bus
            topic   = uri
            payload = bucket.json
            # TODO: Use threads.deferToThread here?
            return self.downstream.publish(topic, payload)

        elif self.scheme == 'influxdb':

            # InfluxDB query wrapper using expression derived from transformation data
            dfq = DataFrameQuery(settings=self.settings, bucket=bucket)

            # Perform query and obtain results as pandas DataFrame
            df = dfq.query()

            # Announce routing information via http response headers
            bucket.request.setHeader('Target-Database', bucket.tdata.database)
            bucket.request.setHeader('Target-Expression', bucket.tdata.expression)
            bucket.request.setHeader('Target-Address-Scheme', self.scheme)
            bucket.request.setHeader('Target-Address-Uri', uri)

            # Database result is empty, send appropriate response
            if df is None or df.empty:
                return self.response_no_results(bucket)


            # DataFrame manipulation

            # Drop some fields from DataFrame as requested
            if 'exclude' in bucket.tdata and bucket.tdata.exclude:
                drop_fields = read_list(bucket.tdata.exclude, empty_elements=False)
                try:
                    df.drop(drop_fields, axis=1, inplace=True)
                except ValueError as ex:
                    log.error(last_error_and_traceback())
                    error_message = u'Error: {type} {message}'.format(type=type(ex), message=ex)
                    return bucket.request.error_response(bucket, error_message=error_message)

            # Use only specified fields from DataFrame as requested
            if 'include' in bucket.tdata and bucket.tdata.include:
                use_fields = read_list(bucket.tdata.include, empty_elements=False)
                use_fields.insert(0, 'time')
                try:
                    df = df.filter(use_fields, axis=1)
                except ValueError as ex:
                    log.error(last_error_and_traceback())
                    error_message = u'Error: {type} {message}'.format(type=type(ex), message=ex)
                    return bucket.request.error_response(bucket, error_message=error_message)

            # Propagate non-null values forward or backward.
            # With time series data, using pad/ffill is extremely common so that the “last known value” is available at every time point.
            # http://pandas.pydata.org/pandas-docs/stable/missing_data.html#filling-missing-values-fillna
            if 'pad' in bucket.tdata and asbool(bucket.tdata.pad):
                df.fillna(method='pad', inplace=True)

            if 'backfill' in bucket.tdata and asbool(bucket.tdata.backfill):
                df.fillna(method='backfill', inplace=True)

            if 'interpolate' in bucket.tdata and asbool(bucket.tdata.interpolate):
                # Performs linear interpolation at missing datapoints,
                # otherwise matplotlib would not plot the sparse data frame.
                # http://pandas.pydata.org/pandas-docs/stable/missing_data.html#interpolation
                df.interpolate(inplace=True)

            if 'sorted' in bucket.tdata and asbool(bucket.tdata.sorted):
                # http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.sort.html
                df.sort(axis='columns', inplace=True)


            # Compute http response from DataFrame, taking designated output format into account
            response = HttpDataFrameResponse(bucket, dataframe=df)

            # Synchronous, the worker-threading is already on the HTTP layer
            return response.render()

            # Asynchronous: Perform computation in separate thread
            d = threads.deferToThread(response.render)
            d.addErrback(handleFailure, bucket.request)
            d.addBoth(bucket.request.write)
            d.addBoth(lambda _: bucket.request.finish())
            return server.NOT_DONE_YET

        else:
            message = 'No target/downstream dispatcher for scheme {scheme}'.format(scheme=self.scheme)
            log.error(message)
            raise KeyError(message)

    def response_no_results(self, bucket):
        """
        Send "404 Not Found" response body in text/plain
        format describing HTTP API query interface.
        """
        # FIXME: Some words about "now-10d" being the default for "from", see influx.py.
        # FIXME: Maybe refactor "now-10d" to transformation machinery to make it accessible from here.
        error_message = u'# 404 Not Found\n#\n'\
                        u'# No data for query expression "{expression}"\n'\
                        u'# Please recognize absolute datetimes are expected to be in ISO 8601 format. '\
                        u'Default is UTC, optionally specify an appropriate timezone offset.\n'.format(expression=bucket.tdata.expression)
        error_message += u'#\n# Examples:\n#\n'\
                         u'#   ?from=2016-06-25T22:00:00.000Z\n'\
                         u'#   ?from=2016-06-26T00:00:00.000%2B02:00    (%2B is "+" urlencoded)\n'\
                         u'#   ?from=now-4h&to=now-2h\n'\
                         u'#   ?from=now-8d5h3m&to=now-6d'
        bucket.request.setResponseCode(http.NOT_FOUND)
        bucket.request.setHeader('Content-Type', 'text/plain; charset=utf-8')
        return error_message.encode('utf-8')




# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import os
import sys
import logging
from pprint import pprint
from binascii import unhexlify, hexlify
from tabulate import tabulate
from kotori.util.common import setup_logging
from kotori.vendor.lst.h2m.util import setup_h2m_structs
from kotori.daq.intercom.c import LibraryAdapter, StructRegistry, StructRegistryByID

logger = logging.getLogger(__name__)


def getting_started(sr):
# create a "cap_r" struct and print attributes with default values
    print '-' * 42
    p = sr.create('struct_cap_r')
    print 'thing: ', p
    print 'length:', p.length
    print 'ID:    ', p.ID

    # create a "cap_r" struct overriding default values and print the relevant attributes
    print '-' * 42
    p = sr.create('struct_cap_r', ID=88)
    print 'thing: ', p
    print 'length:', p.length
    print 'ID:    ', p.ID


    # get lowlevel pyclibrary ctypes backend handle of "struct_program"
    print '-' * 42
    print 'struct_program (schema):'
    struct_program = sr.get('struct_program')
    pprint(struct_program)
    csp = struct_program.obj() #(abc=9)
    print 'thing:      ', csp
    print 'length:     ', csp.length
    print 'ID:         ', csp.ID
    print 'send_ser:   ', csp.send_ser
    print 'cfg_loaded: ', csp.cfg_loaded

    print '-' * 42
    print 'struct_program (instance):'
    print 'thing:      ', csp()
    print 'length:     ', csp().length
    print 'ID:         ', csp().ID


    # get lowlevel pyclibrary ctypes backend handle of "struct_cap_r"
    print '-' * 42
    c = sr.get('struct_cap_r').obj()
    print 'thing: ', c
    print 'length:', c.length
    print 'ID:    ', c.ID

    print 'FIELDS:'
    pprint(c._fields_)
    print 'DEFAULTS:'
    pprint(c._defaults_)


def dump_full(sr):

    """
    s_program = sr.get('struct_program')
    print s_program
    ast = s_program.ast()
    print dir(ast.members)
    print ast.members
    """

    program = sr.create('struct_program')
    sr.pprint(program)

    gps_w = sr.create('struct_gps_w')
    sr.pprint(gps_w)


def fill_and_dump(sr):

    program = sr.create('struct_program')
    d = sr.to_dict(program)
    pprint(list(d.iteritems()))
    print 'dump-1:', hexlify(program._dump_())

    remote = unhexlify('0d99000000000000000000000000')
    program._load_(remote)
    print 'dump-2:', hexlify(program._dump_())

    program.ID = 88
    print 'dump-3:', hexlify(program._dump_())
    d = sr.to_dict(program)

    # https://stackoverflow.com/questions/4301069/any-way-to-properly-pretty-print-ordered-dictionaries-in-python
    # https://bugs.python.org/issue10592
    # https://bugs.python.org/issue7434
    # https://stackoverflow.com/questions/21420243/pretty-printing-ordereddicts-using-pprint
    pprint(dict(d.items()))


def pretend_receive_and_process(sr):

    # let's pretend this is the message received, having ID=2
    payload = unhexlify('05022a0021')

    # decode struct ID from binary data
    message_id = ord(payload[1])

    # look up proper StructAdapter object
    # it should be "struct_cap_r", registered with ID=2
    struct_cap_r = sr.get_by_id(message_id)

    # create instance of struct
    cap_r = struct_cap_r.create()

    # load binary message payload into struct
    cap_r._load_(payload)

    # pretty-print struct content
    #sr.pprint(cap_r)
    #print '=' * 21
    sr.pprint(cap_r, format='tabulate-plain')


def display_schema(sr):
    struct_h2o_w = sr.get('struct_h2o_w')
    struct_h2o_w.print_schema()


def main():

    setup_logging(logging.INFO)

    # initialize "h2m_structs" library
    sr = setup_h2m_structs()

    # call some examples
    #getting_started(sr)
    #dump_full(sr)
    #fill_and_dump(sr)
    #pretend_receive_and_process(sr)
    display_schema(sr)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# (c) 2016 Andreas Motl <andreas.motl@elmyra.de>
from pyramid.config.predicates import RequestMethodPredicate
from pyramid.urldispatch import RoutesMapper
from pyramid.threadlocal import get_current_registry

class PathRoutingEngine(object):
    """
    A simple routing engine for path-based patterns
    based on the powerful Pyramid request router.
    """

    def __init__(self):
        self.mapper = RoutesMapper()

    def add_route(self, name, pattern, methods=[]):
        predicates = []
        for method in methods:
            predicate = RequestMethodPredicate(method.upper(), None)
            predicates.append(predicate)
        self.mapper.connect(name, pattern, predicates=predicates)

    def match(self, method, path):
        #print 'PathRoutingEngine attempt to match path:', path
        request = self._getRequest(attributes={'method': method}, environ={'PATH_INFO': path})
        result = self.mapper(request)
        if result['route']:
            #print 'PathRoutingEngine matched result:       ', result
            return result

    def _getRequest(self, attributes, environ):
        # from pyramid.tests.test_urldispatch
        environ_default = {'SERVER_NAME':'localhost',
                           'wsgi.url_scheme':'http'}
        environ.update(environ_default)

        request = DummyRequest(attributes, environ)
        reg = get_current_registry()
        request.registry = reg
        return request

class DummyRequest(object):
    def __init__(self, attributes, environ):
        self.method  = None
        self.environ = environ
        for name, value in attributes.iteritems():
            setattr(self, name, value)


if __name__ == '__main__':
    router = PathRoutingEngine()
    pattern = '/foo/bar/{resource}'
    router.add_route(pattern, pattern)
    print 'match:   ', router.match('GET', '/foo/bar/entity-1')
    print 'mismatch:', router.match('GET', '/hello/world')



# -*- coding: utf-8 -*-
# (c) 2016 Andreas Motl <andreas.motl@elmyra.de>
from twisted.logger import Logger, LogLevel
log = Logger()

try:
    import pandas
    from pandas.tslib import Timedelta
except ImportError:
    log.failure('Dataframe functions not available, please install "pandas".', level=LogLevel.warn)


def dataframe_index_to_column(df, column):
    """
    Copy DataFrame index column to real data column.
    """
    dt = df.index
    df[column] = dt
    #df.reset_index(drop=True, inplace=True)
    return df

def dataframe_wide_to_long_indexed(df, column):
    """
    Convert DataFrame from wide to long format using specified column as index column,
    followed by indexing the DataFrame on the very same column and finally sorting it.

    See also:

    - http://pandas.pydata.org/pandas-docs/stable/reshaping.html#reshaping-by-melt
    - http://stackoverflow.com/questions/17688155/complicated-for-me-reshaping-from-wide-to-long-in-pandas
    """
    df = pandas.melt(df, id_vars=column).dropna()
    df = dataframe_index_and_sort(df, column)
    return df

def dataframe_index_and_sort(df, column):
    """
    Index and sort DataFrame on specified column.
    """
    df = df.set_index([column])
    df = df.sort_index()
    return df

def matplotlib_locator_formatter(timedelta, span=1):
    """
    Compute appropriate locator and formatter for renderers
    based on matplotlib, depending on designated time span.
    """
    from matplotlib.dates import date_ticker_factory, DateFormatter
    locator, formatter = date_ticker_factory(span)

    # http://pandas.pydata.org/pandas-docs/stable/timedeltas.html
    # https://stackoverflow.com/questions/16103238/pandas-timedelta-in-days
    is_macro = timedelta <= Timedelta(days=1)
    is_supermacro = timedelta <= Timedelta(minutes=5)

    if is_macro:
        #formatter = DateFormatter(fmt='%H:%M:%S.%f')
        formatter = DateFormatter(fmt='%H:%M')

    if is_supermacro:
        formatter = DateFormatter(fmt='%H:%M:%S')

        # Formatter overrides
        #if formatter.fmt == '%H:%M\n%b %d':
        #    formatter = DateFormatter(fmt='%Y-%m-%d %H:%M')

    # Labs
    #from matplotlib.dates import AutoDateLocator, AutoDateFormatter, HOURLY
    #locator = AutoDateLocator(maxticks=7)
    #locator.autoscale()
    #locator.intervald[HOURLY] = [5]
    #formatter = AutoDateFormatter(breaks)
    #formatter = date_format('%Y-%m-%d\n%H:%M')

    # Default building blocks
    #from matplotlib.dates import AutoDateFormatter, AutoDateLocator
    #locator = AutoDateLocator()
    #formatter = AutoDateFormatter(locator)

    return locator, formatter




# -*- coding: utf-8 -*-
# (c) 2017,2018 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017,2018 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3

# This program has moved to https://github.com/hiveeyes/luftdatenpumpe
# Please update your links and bookmarks.

if __name__ == '__main__':
    print('This program has moved to https://github.com/hiveeyes/luftdatenpumpe')



# -*- coding: utf-8 -*-
"""
Kotori is a data acquisition, routing and graphing toolkit
Copyright (C) 2014-2019 Andreas Motl, <andreas@getkotori.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import sys
from bunch import Bunch
from docopt import docopt
from pkgutil import extend_path
from twisted.internet import reactor
from twisted.logger import Logger, LogLevel
from kotori.core import APP_NAME, KotoriBootloader
from kotori.frontend.server import boot_frontend
from kotori.util.common import setup_logging
from kotori.util.logger import startLogging
from kotori.util.configuration import get_configuration, get_configuration_file, apply_default_settings

__path__ = extend_path(__path__, __name__)

__doc__ = APP_NAME + """

Usage:
  kotori [--config etc/development.ini] [--debug-io] [--debug-mqtt] [--debug-mqtt-driver] [--debug-influx] [--debug] [--debug-vendor vendor42,vendor-xyz]
  kotori --version
  kotori (-h | --help)

Options:
  --config etc/kotori.ini   Start Kotori with configuration file
  --version                 Show version information
  --debug-io                Enable debug messages for IO subsystem
  --debug-mqtt              Enable debug messages for MQTT
  --debug-mqtt-driver       Enable debug messages for MQTT driver
  --debug-influx            Enable debug messages for InfluxDB
  --debug                   Generic debug flag passed down to other subsystems
  --debug-vendor vendor42   Debug flag passed down to given vendor subsystems
  -h --help                 Show this screen

"""

log = Logger()

def run():

    setup_logging()
    log.info(u'Starting ' + APP_NAME)

    # Read commandline options
    # TODO: Do it the Twisted way
    options = docopt(__doc__, version=APP_NAME)

    # Read settings from configuration file
    configfile = get_configuration_file(options['--config'])
    log.info("Using configuration file {configfile}", configfile=configfile)
    settings = get_configuration(configfile)

    # Apply default settings
    apply_default_settings(settings)

    # Merge command line options into settings
    settings.setdefault('options', Bunch())
    for key, value in options.iteritems():
        key = key.lstrip(u'--')
        key = key.replace(u'-', u'_')
        settings.options[key] = value

    # Setup the logging subsystem
    log_level = 'info'
    if settings.options.debug:
        log_level = 'debug'
    startLogging(settings, stream=sys.stderr, level=LogLevel.levelWithName(log_level))

    # Boot all enabled applications and vendors
    loader = KotoriBootloader(settings=settings)
    loader.boot_applications()
    loader.boot_vendors()

    # Boot web configuration GUI
    if 'config-web' in settings:
        boot_frontend(settings, debug=settings.options.debug)

    # Enter Twisted reactor loop
    reactor.run()


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
# (c) 2016 Andreas Motl <andreas.motl@elmyra.de>
from twisted.web.template import XMLString, Element, renderer, tags, CDATA
#from dyplot.dygraphs import Dygraphs

class GenericPage(Element):
    """
    Base class for HTML-based widgets.
    """

    @renderer
    def fill_slots(self, request, tag):

        slotdata = self.__dict__.copy()

        # Pass data to template verbatim without applying htmlentity conversion
        slotdata['data_uri'] = CDATA(slotdata['data_uri'])

        tag.fillSlots(**slotdata)
        return tag

    @renderer
    def header(self, request, tag):
        tdata = self.bucket.tdata
        title = 'Address: {network} » {gateway} » {node}'.format(**dict(tdata))
        return tag(tags.b(title))

    @renderer
    def footer_left(self, request, tag):
        return tag(tags.p('Query expression (times are UTC): ', self.bucket.tdata.expression))

    @renderer
    def footer_right(self, request, tag):
        """
        '(c) 2014-2016 ',
        tags.a('Open Hive', href='http://open-hive.org/'), ' and ',
        tags.a('Hiveeyes', href='https://hiveeyes.org/docs/system/'), '. ',
        """
        return tag(tags.p(
            'Powered by ',
            tags.a('Kotori', href='https://getkotori.org/'), ', ',
            tags.a('InfluxDB', href='https://github.com/influxdata/influxdb'), ', ',
            tags.a('dygraphs', href='http://dygraphs.com/'), ' and ',
            tags.a('DataTables', href='https://datatables.net/'), '.'
        ))


class DygraphsPage(GenericPage):
    """
    dygraphs plugin, see http://dygraphs.com/
    """

    loader = XMLString("""
        <html xmlns:t="http://twistedmatrix.com/ns/twisted.web.template/0.1" t:render="fill_slots">
            <head>
                <script src="//cdnjs.cloudflare.com/ajax/libs/dygraph/1.1.1/dygraph-combined.js"></script>
                <style type="text/css" media="screen">
                    body {
                        font-family: Arial, sans-serif;
                        font-size: larger;
                    }
                    h1 {
                        font-size: large;
                    }
                    .footer {
                        padding-top: 1em;
                        font-size: x-small;
                    }
                    .footer.left {
                        float: left;
                    }
                    .footer.right {
                        float: right;
                    }
                    .clearfix {
                        clear: both;
                    }
                    .linegraph {
                        width: 100%;
                        height: 25em;
                    }
                </style>
            </head>
            <body>
                <h1 t:render="header"/>
                <p style="font-size: smaller">
                    <strong>Zoom:</strong> Click and Drag |
                    <strong>Pan:</strong> Shift-Click and Drag |
                    <strong>Unzoom:</strong> Double-Click
                </p>
                <div id="linegraph" class="linegraph"/>
                <script type="text/javascript">
                    var element = document.getElementById('linegraph');
                    var data_uri = '<t:slot name="data_uri" />'.replace('<![CDATA[', '').replace(']]>', '');
                    new Dygraph(element, data_uri, {
                      legend: 'always',
                    });
                </script>
                <div class="footer left" t:render="footer_left"/>
                <div class="footer right" t:render="footer_right"/>
                <div class="clearfix"/>
            </body>
        </html>
    """)

    def __init__(self, data_uri=None, bucket=None):
        self.data_uri = data_uri
        self.bucket = bucket


class DatatablesPage(GenericPage):
    """
    DataTables plugin, see https://datatables.net/
    """

    loader = XMLString("""
        <html xmlns:t="http://twistedmatrix.com/ns/twisted.web.template/0.1" t:render="fill_slots">
            <head>
                <link rel="stylesheet" type="text/css" href="//cdn.datatables.net/u/dt/jq-2.2.3,jszip-2.5.0,pdfmake-0.1.18,dt-1.10.12,b-1.2.1,b-colvis-1.2.1,b-html5-1.2.1,b-print-1.2.1,cr-1.3.2,fc-3.2.2,fh-3.1.2,kt-2.1.2,r-2.1.0,rr-1.1.2,sc-1.4.2,se-1.2.0/datatables.min.css" />
                <script src="//cdn.datatables.net/u/dt/jq-2.2.3,jszip-2.5.0,pdfmake-0.1.18,dt-1.10.12,b-1.2.1,b-colvis-1.2.1,b-html5-1.2.1,b-print-1.2.1,cr-1.3.2,fc-3.2.2,fh-3.1.2,kt-2.1.2,r-2.1.0,rr-1.1.2,sc-1.4.2,se-1.2.0/datatables.min.js"></script>
                <style type="text/css" media="screen">
                    body {
                        font-family: Arial, sans-serif;
                        font-size2: larger;
                    }
                    h1 {
                        font-size: large;
                    }
                    .footer {
                        padding-top: 1em;
                        font-size: x-small;
                    }
                    .footer.left {
                        float: left;
                    }
                    .footer.right {
                        float: right;
                    }
                    .clearfix {
                        clear: both;
                    }
                </style>
            </head>
            <body>
                <h1 t:render="header"/>
                <div id="dataframe" class="display" style="width: 100%"/>
                <table id="example" class="display" cellspacing="0" width="100%">
                </table>
                <script type="text/javascript">
                    $(document).ready(function(){
                        var data_uri = '<t:slot name="data_uri" />'.replace('<![CDATA[', '').replace(']]>', '');
                        $('#dataframe').load(data_uri, function() {
                            var table = $('.dataframe');
                            table.attr('border', '0');
                            table.DataTable({
                                buttons: [
                                    //'copy', 'csv', 'excel', 'pdf', 'print'
                                    'copy', 'csv', 'excel', 'print'
                                ]
                            });
                            table.attr('class', 'display dataTable');
                        });
                        console.log('READY.');

                    });
                </script>
                <div class="footer left" t:render="footer_left"/>
                <div class="footer right" t:render="footer_right"/>
                <div class="clearfix"/>
            </body>
        </html>
    """)

    def __init__(self, data_uri=None, bucket=None):
        self.data_uri = data_uri
        self.bucket = bucket




# Geographical helper functions for nmea_info.py and friends
#
# Helps with geographic functions, including:
#  Lat+Long+Height -> XYZ
#  XYZ -> Lat+Long+Height
#  Lat+Long -> other Lat+Long (Helmert Transform)
#  Lat+Long -> easting/northing (OS GB+IE Only)
#  easting/northing -> Lat+Long (OS GB+IE Only)
#  OS easting/northing -> OS 6 figure ref
#
# See http://gagravarr.org/code/ for updates and information
#
# GPL
#
# Nick Burch - v0.06 (30/05/2007)

import math

# For each co-ordinate system we do, what are the A, B and E2 values?
# List is A, B, E^2 (E^2 calculated after)
abe_values = {
	'wgs84': [ 6378137.0, 6356752.3141, -1 ],
	'osgb' : [ 6377563.396, 6356256.91, -1 ],
	'osie' : [ 6377340.189, 6356034.447, -1 ]
}

# The earth's radius, in meters, as taken from an average of the WGS84
#  a and b parameters (should be close enough)
earths_radius = (abe_values['wgs84'][0] + abe_values['wgs84'][1]) / 2.0

# Calculate the E2 values
for system in abe_values.keys():
	a = abe_values[system][0]
	b = abe_values[system][1]
	e2 = (a*a - b*b) / (a*a)
	abe_values[system][2] = e2

# For each co-ordinate system we can translate between, what are
#  the tx, ty, tz, s, rx, ry and rz values?
# List is tx, ty, tz, s, rx, ry, rz
transform_values = {
	'wgs84_to_osgb' : [ -446.448, 125.157, -542.060, 
							20.4894 / 1000.0 / 1000.0, # given as ppm
							-0.1502 / 206265.0, # given as seconds of arc
							-0.2470 / 206265.0, # given as seconds of arc
							-0.8421 / 206265.0  # given as seconds of arc
					],
	'wgs84_to_osie' : [ -482.530, 130.596, -564.557, 
							-8.1500 / 1000.0 / 1000.0, # given as ppm
							-1.0420 / 206265.0, # given as seconds of arc
							-0.2140 / 206265.0, # given as seconds of arc
							-0.6310 / 206265.0  # given as seconds of arc
					],
	'itrs2000_to_etrs89' : [ 0.054, 0.051, -0.048, 0,
							 0.000081 / 206265.0, # given as seconds of arc
							 0.00049 / 206265.0, # given as seconds of arc
							 0.000792 / 206265.0  # given as seconds of arc
					]
}

# Calculate reverse transforms
for systems in [('wgs84','osgb'), ('wgs84','osie'), ('itrs2000','etrs89')]:
	fs = systems[0] + "_to_" + systems[1]
	rs = systems[1] + "_to_" + systems[0]
	ra = []
	for val in transform_values[fs]:
		ra.append(-1.0 * val)
	transform_values[rs] = ra

# Easting and Northin system values, for the systems we work with.
# List is n0, e0, F0, theta0 and landa0
en_values = {
	'osgb' : [ -100000.0, 400000.0, 0.9996012717,
					49.0 /360.0 *2.0*math.pi,
					-2.0 /360.0 *2.0*math.pi
			 ],
	'osie' : [ 250000.0, 200000.0, 1.000035,
					53.5 /360.0 *2.0*math.pi,
					-8.0 /360.0 *2.0*math.pi
			 ]
}

# Cassini Projection Origins
# List is lat (rad), long (rad), false easting, false northing
cassini_values = {
	'osgb' : [ (53.0 + (13.0 / 60.0) + (17.274 / 3600.0)) /360.0 *2.0*math.pi,
	           -(2.0 + (41.0 / 60.0) +  (3.562 / 3600.0)) /360.0 *2.0*math.pi,
	           0, 0 ]				 
}

# How many feet to the meter
feet_per_meter = 1.0 / 0.3048007491   # 3.28083

##############################################################
#         OS GB Specific Helpers for Generic Methods         #
##############################################################

def turn_wgs84_into_osgb36(lat_dec,long_dec,height):
	"""See http://www.gps.gov.uk/guide6.asp#6.2 and http://www.gps.gov.uk/guide6.asp#6.6 for the calculations, and http://www.posc.org/Epicentre.2_2/DataModel/ExamplesofUsage/eu_cs34h.html for some background."""

	wgs84_xyz = turn_llh_into_xyz(lat_dec,long_dec,height,'wgs84')

	osgb_xyz = turn_xyz_into_other_xyz(
					wgs84_xyz[0],wgs84_xyz[1],wgs84_xyz[2],'wgs84','osgb')

	osgb_latlong = turn_xyz_into_llh(
					osgb_xyz[0],osgb_xyz[1],osgb_xyz[2],'osgb')
	return osgb_latlong

def turn_osgb36_into_wgs84(lat_dec,long_dec,height):
	"""See http://www.gps.gov.uk/guide6.asp#6.2 and http://www.gps.gov.uk/guide6.asp#6.6 for the calculations, and http://www.posc.org/Epicentre.2_2/DataModel/ExamplesofUsage/eu_cs34h.html for some background."""

	osgb_xyz = turn_llh_into_xyz(lat_dec,long_dec,height,'osgb')

	wgs84_xyz = turn_xyz_into_other_xyz(
					osgb_xyz[0],osgb_xyz[1],osgb_xyz[2],'osgb','wgs84')

	wgs84_latlong = turn_xyz_into_llh(
					wgs84_xyz[0],wgs84_xyz[1],wgs84_xyz[2],'wgs84')

	return wgs84_latlong

def turn_osgb36_into_eastingnorthing(lat_dec,long_dec):
	"""Turn OSGB36 (decimal) lat/long values into OS easting and northing values."""
	return turn_latlong_into_eastingnorthing(lat_dec,long_dec,'osgb')

def turn_eastingnorthing_into_osgb36(easting,northing):
	"""Turn OSGB36 easting and northing values into (decimal) lat/long values inOSGB36."""
	return turn_eastingnorthing_into_latlong(easting,northing,'osgb')

##############################################################
#         OS IE Specific Helpers for Generic Methods         #
##############################################################

def turn_wgs84_into_osie36(lat_dec,long_dec,height):
	"""As per turn_wgs84_into_osgb36, but for Irish grid"""

	wgs84_xyz = turn_llh_into_xyz(lat_dec,long_dec,height,'wgs84')

	osie_xyz = turn_xyz_into_other_xyz(
					wgs84_xyz[0],wgs84_xyz[1],wgs84_xyz[2],'wgs84','osie')

	osie_latlong = turn_xyz_into_llh(
					osie_xyz[0],osie_xyz[1],osie_xyz[2],'osie')
	return osie_latlong

def turn_osie36_into_wgs84(lat_dec,long_dec,height):
	"""As per turn_osgb36_into_wgs84, but for Irish grid"""

	osie_xyz = turn_llh_into_xyz(lat_dec,long_dec,height,'osie')

	wgs84_xyz = turn_xyz_into_other_xyz(
					osie_xyz[0],osie_xyz[1],osie_xyz[2],'osie','wgs84')

	wgs84_latlong = turn_xyz_into_llh(
					wgs84_xyz[0],wgs84_xyz[1],wgs84_xyz[2],'wgs84')

	return wgs84_latlong

def turn_osie36_into_eastingnorthing(lat_dec,long_dec):
	"""Turn OSIE36 (decimal) lat/long values into OS IE easting and northing values."""
	return turn_latlong_into_eastingnorthing(lat_dec,long_dec,'osie')

def turn_eastingnorthing_into_osie36(easting,northing):
	"""Turn OSIE36 easting and northing values into (decimal) lat/long values inOSIE36."""
	return turn_eastingnorthing_into_latlong(easting,northing,'osie')

##############################################################
#             Generic Transform Functions                    #
##############################################################

def turn_llh_into_xyz(lat_dec,long_dec,height,system):
	"""Convert Lat, Long and Height into 3D Cartesian x,y,z
       See http://www.ordnancesurvey.co.uk/gps/docs/convertingcoordinates3D.pdf"""

	a = abe_values[system][0]
	b = abe_values[system][1]
	e2 = abe_values[system][2]

	theta = float(lat_dec)  / 360.0 * 2.0 * math.pi
	landa = float(long_dec) / 360.0 * 2.0 * math.pi
	height = float(height)

	v = a / math.sqrt( 1.0 - e2 * (math.sin(theta) * math.sin(theta)) )
	x = (v + height) * math.cos(theta) * math.cos(landa)
	y = (v + height) * math.cos(theta) * math.sin(landa)
	z = ( (1.0 - e2) * v + height ) * math.sin(theta)

	return [x,y,z]

def turn_xyz_into_llh(x,y,z,system):
	"""Convert 3D Cartesian x,y,z into Lat, Long and Height
       See http://www.ordnancesurvey.co.uk/gps/docs/convertingcoordinates3D.pdf"""

	a = abe_values[system][0]
	b = abe_values[system][1]
	e2 = abe_values[system][2]

	p = math.sqrt(x*x + y*y)

	long = math.atan(y/x)
	lat_init = math.atan( z / (p * (1.0 - e2)) )
	v = a / math.sqrt( 1.0 - e2 * (math.sin(lat_init) * math.sin(lat_init)) )
	lat = math.atan( (z + e2*v*math.sin(lat_init)) / p )

	height = (p / math.cos(lat)) - v # Ignore if a bit out

	# Turn from radians back into degrees
	long = long / 2 / math.pi * 360
	lat = lat / 2 / math.pi * 360

	return [lat,long,height]

def turn_xyz_into_other_xyz(old_x,old_y,old_z,from_scheme,to_scheme):
	"""Helmert Transformation between one lat+long system and another
See http://www.ordnancesurvey.co.uk/oswebsite/gps/information/coordinatesystemsinfo/guidecontents/guide6.html for the calculations, and http://www.movable-type.co.uk/scripts/LatLongConvertCoords.html for a friendlier version with examples"""

	transform = from_scheme + "_to_" + to_scheme
	tx = transform_values[transform][0]
	ty = transform_values[transform][1]
	tz = transform_values[transform][2]
	s =  transform_values[transform][3]
	rx = transform_values[transform][4]
	ry = transform_values[transform][5]
	rz = transform_values[transform][6]

	# Do the transform
	new_x = tx + ((1.0+s) * old_x) + (-rz * old_y) + (ry * old_z)
	new_y = ty + (rz * old_x) + ((1.0+s) * old_y) + (-rx * old_z)
	new_z = tz + (-ry * old_x) + (rx * old_y) + ((1.0+s) * old_z)

	return [new_x,new_y,new_z]

def calculate_distance_and_bearing(from_lat_dec,from_long_dec,to_lat_dec,to_long_dec):
	"""Uses the spherical law of cosines to calculate the distance and bearing between two positions"""

	# Turn them all into radians
	from_theta = float(from_lat_dec)  / 360.0 * 2.0 * math.pi
	from_landa = float(from_long_dec) / 360.0 * 2.0 * math.pi
	to_theta = float(to_lat_dec)  / 360.0 * 2.0 * math.pi
	to_landa = float(to_long_dec) / 360.0 * 2.0 * math.pi

	d = math.acos(
			math.sin(from_theta) * math.sin(to_theta) +
			math.cos(from_theta) * math.cos(to_theta) * math.cos(to_landa-from_landa)
		) * earths_radius

	bearing = math.atan2(
			math.sin(to_landa-from_landa) * math.cos(to_theta),
			math.cos(from_theta) * math.sin(to_theta) -
			math.sin(from_theta) * math.cos(to_theta) * math.cos(to_landa-from_landa)
		)
	bearing = bearing / 2.0 / math.pi * 360.0

	return [d,bearing]

##############################################################
#            Easting/Northing Transform Methods              #
##############################################################

def turn_latlong_into_eastingnorthing(lat_dec,long_dec,scheme):
	"""Turn OSGB36 or OSIE36 (decimal) lat/long values into OS easting and northing values. See http://www.ordnancesurvey.co.uk/oswebsite/gps/information/coordinatesystemsinfo/guidecontents/guide7.html for the calculations, and http://www.posc.org/Epicentre.2_2/DataModel/ExamplesofUsage/eu_cs34h.html for some background."""

	n0 = en_values[scheme][0]
	e0 = en_values[scheme][1]
	f0 = en_values[scheme][2]

	theta0 = en_values[scheme][3]
	landa0 = en_values[scheme][4]

	a = abe_values[scheme][0]
	b = abe_values[scheme][1]
	e2 = abe_values[scheme][2]

	theta = float(lat_dec)  /360.0 *2.0*math.pi
	landa = float(long_dec) /360.0 *2.0*math.pi

	n = (a-b) / (a+b)
	v = a * f0 * math.pow( (1 - e2 * math.sin(theta)*math.sin(theta)), -0.5 )
	ro = a * f0 * (1 - e2) * math.pow( (1 - e2 * math.sin(theta)*math.sin(theta)), -1.5 )
	nu2 = v/ro - 1

	M = b * f0 * ( \
		(1.0 + n + 5.0/4.0 *n*n + 5.0/4.0 *n*n*n) * (theta-theta0) - \
		(3.0*n + 3.0*n*n + 21.0/8.0 *n*n*n) *math.sin(theta-theta0) *math.cos(theta+theta0) + \
		(15.0/8.0*n*n + 15.0/8.0*n*n*n) *math.sin(2.0*(theta-theta0)) *math.cos(2.0*(theta+theta0)) - \
		35.0/24.0*n*n*n *math.sin(3.0*(theta-theta0)) *math.cos(3.0*(theta+theta0)) \
	)

	I = M + n0
	II = v/2.0 * math.sin(theta) * math.cos(theta)
	III = v/24.0 * math.sin(theta) * math.pow( math.cos(theta),3 ) * \
		(5.0 - math.pow(math.tan(theta),2) + 9.0*nu2)
	IIIa = v/720.0 * math.sin(theta) * math.pow( math.cos(theta),5 ) * \
		( 61.0 - 58.0 *math.pow(math.tan(theta),2) + math.pow(math.tan(theta),4) )
	IV = v * math.cos(theta)
	V = v/6.0 * math.pow( math.cos(theta),3 ) * \
		( v/ro - math.pow(math.tan(theta),2) )
	VI = v/120.0 * math.pow(math.cos(theta),5) * \
		( 5.0 - 18.0 *math.pow(math.tan(theta),2) + \
		math.pow(math.tan(theta),4) + 14.0*nu2 - \
		58.0 * math.pow(math.tan(theta),2)*nu2 )

	northing = I + II*math.pow(landa-landa0,2) + \
		III*math.pow(landa-landa0,4) + \
		IIIa*math.pow(landa-landa0,6)
	easting = e0 + IV*(landa-landa0) + V*math.pow(landa-landa0,3) + \
		VI*math.pow(landa-landa0,5)
	
	return (easting,northing)

def turn_eastingnorthing_into_latlong(easting,northing,scheme):
	"""Turn OSGB36 or OSIE36 easting and northing values into (decimal) lat/long values in OSGB36 / OSIE36. See http://www.ordnancesurvey.co.uk/oswebsite/gps/information/coordinatesystemsinfo/guidecontents/guide7.html for the calculations, and http://www.posc.org/Epicentre.2_2/DataModel/ExamplesofUsage/eu_cs34h.html for some background."""

	n0 = en_values[scheme][0]
	e0 = en_values[scheme][1]
	f0 = en_values[scheme][2]

	theta0 = en_values[scheme][3]
	landa0 = en_values[scheme][4]

	a = abe_values[scheme][0]
	b = abe_values[scheme][1]
	e2 = abe_values[scheme][2]

	n = (a-b) / (a+b)

	# Prepare to iterate
	M = 0
	theta = theta0
	# Iterate, 4 times should be enough
	for i in range(4):
		theta = ((northing - n0 - M) / (a * f0)) + theta
		M = b * f0 * ( \
			(1.0 + n + 5.0/4.0 *n*n + 5.0/4.0 *n*n*n) * (theta-theta0) - \
			(3.0*n + 3.0*n*n + 21.0/8.0 *n*n*n) *math.sin(theta-theta0) *math.cos(theta+theta0) + \
			(15.0/8.0*n*n + 15.0/8.0*n*n*n) *math.sin(2.0*(theta-theta0)) *math.cos(2.0*(theta+theta0)) - \
			35.0/24.0*n*n*n *math.sin(3.0*(theta-theta0)) *math.cos(3.0*(theta+theta0)) \
		)

	# Compute intermediate values
	v = a * f0 * math.pow( (1 - e2 * math.sin(theta)*math.sin(theta)), -0.5 )
	ro = a * f0 * (1 - e2) * math.pow( (1 - e2 * math.sin(theta)*math.sin(theta)), -1.5 )
	nu2 = v/ro - 1
	tantheta2 = math.pow(math.tan(theta),2)

	VII = math.tan(theta) / (2 * ro * v)
	VIII = math.tan(theta) / (24 * ro * math.pow(v,3)) \
			* (5 + 3 * tantheta2 + nu2 - \
			   9 * tantheta2 *  nu2 )
	IX = math.tan(theta) / (720 * ro * math.pow(v,5)) \
			* (61 + 90 * tantheta2 + 45 * tantheta2 * tantheta2)
	X = 1 / (math.cos(theta) * v)
	XI = 1 / (math.cos(theta) * 6 * math.pow(v,3)) \
			* (v/ro + 2*tantheta2)
	XII = 1 / (math.cos(theta) * 120 * math.pow(v,5)) \
			* (5 + 28 * tantheta2 + 24 * tantheta2 * tantheta2)
	XIIa = 1 / (math.cos(theta) * 5040 * math.pow(v,7)) \
			* (61 + 662 * tantheta2 + 1320 * tantheta2 * tantheta2 \
				+ 720 * tantheta2 * tantheta2 * tantheta2)

	lat_rad = theta - VII * math.pow((easting-e0),2) \
				+ VIII * math.pow((easting-e0),4) \
				- IX * math.pow((easting-e0),6)
	long_rad = landa0 + X * (easting-e0) \
				- XI * math.pow((easting-e0),3) \
				+ XII * math.pow((easting-e0),5) \
				- XIIa * math.pow((easting-e0),7)

	lat = lat_rad / 2.0 / math.pi * 360.0
	long = long_rad / 2.0 / math.pi * 360.0

	return (lat,long)

##############################################################
#         Cassini Easting/Northing Transform Methods         #
##############################################################

def turn_latlong_into_cassini_en(lat_dec,long_dec,scheme):
	"""Latitude and Longitude, into Cassini-Soldner easting and northing co-ordinates, in the given scheme. See http://www.posc.org/Epicentre.2_2/DataModel/ExamplesofUsage/eu_cs34g.html for details of the calculation used"""
	
	a = abe_values[scheme][0]
	b = abe_values[scheme][1]
	e2 = abe_values[scheme][2]

	e4 = e2 * e2
	e6 = e2 * e2 * e2

	theta = float(lat_dec)  /360.0 *2.0*math.pi
	landa = float(long_dec) /360.0 *2.0*math.pi

	theta0 = cassini_values[scheme][0]
	landa0 = cassini_values[scheme][1]
	false_easting = cassini_values[scheme][2]
	false_northing = cassini_values[scheme][3]

	# Compute intermediate values
	A = (landa - landa0) * math.cos(theta)
	T = math.tan(theta) * math.tan(theta)
	C = e2 / (1.0 - e2) * math.cos(theta) * math.cos(theta)
	v = a / math.sqrt( 1 - (e2 * math.sin(theta) * math.sin(theta)) )

	A2 = A ** 2
	A3 = A ** 3
	A4 = A ** 4
	A5 = A ** 5

	# And M, which is how far along the meridian our latitude is from the origin
	def makeM(picked_theta):
		return a * (
			(1.0 - e2/4.0 - 3.0*e4/64.0 - 5.0*e6/256.0) * picked_theta
		  - (3.0*e2/8.0 + 3.0*e4/32.0 + 45.0*e6/1024.0) * math.sin(2.0*picked_theta)
		  + (15.0*e4/256.0 + 45.0*e6/1024.0) * math.sin(4.0*picked_theta)
		  - (35.0*e6/3072.0) * math.sin(6.0*picked_theta)
	    )
	M = makeM(theta)
	M0 = makeM(theta0)

	# Now calculate
	easting = false_easting + v * (
				A - T * A3 / 6.0 - (8.0 - T + 8.0*C) * T * A5 / 120.0 )
	northing = false_northing + M - M0 + v * math.tan(theta) * (
				A2 / 2.0 + (5.0 - T + 6.0*C) * A4 / 24.0 )

	return (easting,northing)

def turn_cassini_en_into_latlong(easting,northing,scheme):
	"""Cassini-Soldner easting and northing, into Latitude and Longitude, in the given scheme. See http://www.posc.org/Epicentre.2_2/DataModel/ExamplesofUsage/eu_cs34g.html for details of the calculation used"""
	
	a = abe_values[scheme][0]
	b = abe_values[scheme][1]
	e2 = abe_values[scheme][2]

	e4 = e2 * e2
	e6 = e2 * e2 * e2

	theta0 = cassini_values[scheme][0]
	landa0 = cassini_values[scheme][1]
	false_easting = cassini_values[scheme][2]
	false_northing = cassini_values[scheme][3]

	def makeM(picked_theta):
		return a * (
			(1.0 - e2/4.0 - 3.0*e4/64.0 - 5.0*e6/256.0) * picked_theta
		  - (3.0*e2/8.0 + 3.0*e4/32.0 + 45.0*e6/1024.0) * math.sin(2.0*picked_theta)
		  + (15.0*e4/256.0 + 45.0*e6/1024.0) * math.sin(4.0*picked_theta)
		  - (35.0*e6/3072.0) * math.sin(6.0*picked_theta)
	    )

	# Compute first batch of intermediate values
	M1 = makeM(theta0) + (northing - false_northing)
	mu1 = M1 / (a * (1.0 - e2/4.0 - 3.0*e4/64.0 - 5.0*e6/256.0) )
	e1 = (1 - ((1-e2) ** 0.5)) / (1 + ((1-e2) ** 0.5))

	e1_2 = e1 ** 2
	e1_3 = e1 ** 3
	e1_4 = e1 ** 4

	# Now compute theta1 at T1
	theta1 = mu1 + (
        + (3.0*e1 / 2.0 - 27.0*e1_3 / 32.0) * math.sin(2.0*mu1)
        + (21.0*e1_2 / 16.0 - 55.0*e1_4 / 32.0) * math.sin(4.0*mu1)
        + (151.0*e1_3 / 96.0) * math.sin(6.0*mu1)
        + (1097.0*e1_4 / 512.0) * math.sin(8.0*mu1) 
	)
	T1 = (math.tan(theta1)) ** 2

	# Now we can find v1, ro1 and D
	v1 = a / math.sqrt( 1.0 - (e2 * math.sin(theta1) * math.sin(theta1)) )
	ro1 = a * (1 - e2) / ((1 - e2 * math.sin(theta1) * math.sin(theta1)) ** 1.5)
	D = (easting - false_easting) / v1

	# And finally the lat and long
	lat = theta1 - (v1 * math.tan(theta1)) / ro1 * (
			D*D/2.0 - (1.0 + 3.0 * T1) * ( (D**4) / 24.0 ) )
	long = landa0 + (
				D - T1 * (D**3) / 3.0 + (1 + 3.0 * T1) * T1 * (D**5) / 15.0
			) / math.cos(theta1)

	# Now make decimal versions
	lat_dec = lat * 360.0 / 2.0 / math.pi
	long_dec = long * 360.0 / 2.0 / math.pi

	return (lat_dec,long_dec)

##############################################################
#             OS Specific Methods Follow                     #
##############################################################

def turn_easting_northing_into_six_fig(easting,northing):
	"""Turn OS easting and northing values into the six figure OS grid refecence. See http://www.jstott.me.uk/jscoord/"""
	first_letter = ""
	second_letter = ""

	easting = int(easting)
	northing = int(northing)

	# Get the 100 km part
	hkm_east = int( math.floor(easting / 100000.0) )
	hkm_north = int( math.floor(northing / 100000.0) )
	if hkm_north < 5:
		if hkm_east < 5:
			first_letter = "S"
		else:
			first_letter = "T"
	elif hkm_north < 10:
		if hkm_east < 5:
			first_letter = "N"
		else:
			first_letter = "O"
	else:
		first_letter = "H"

	# Get the 10km part
	index = 65 + ((4 - (hkm_north % 5)) * 5) + (hkm_east % 5)
	ti = index
	if index >= 73:
		index += 1
	second_letter = chr(index)

	# Get digits 2-4 on easting and northing
	e = math.floor( (easting  - (100000.0 * hkm_east))  / 100.0)
	n = math.floor( (northing - (100000.0 * hkm_north)) / 100.0)
	e = "%03d" % e
	n = "%03d" % n

	return first_letter + second_letter + e + n



# -*- coding: utf-8 -*-
# (c) 2016-2017 Andreas Motl <andreas@getkotori.org>
import threading
from copy import deepcopy
from urlparse import urlparse
from bunch import Bunch, bunchify
from collections import defaultdict
from pyramid.settings import asbool
from twisted.internet import threads
from twisted.web import http, server
from twisted.logger import Logger
from twisted.application.service import MultiService
from kotori.util.configuration import read_list
from kotori.daq.services import RootService, MultiServiceMixin
from kotori.io.protocol.forwarder import ForwarderAddress
from kotori.io.protocol.http import HttpServerService
from kotori.firmware.builder import FirmwareBuilder

"""
Todo
====
- [o] Use threads for running concurrent build operations, but
  safeguard the compilation process using locks around "build_firmware".
"""

log = Logger()

class FirmwareBuilderApplication(RootService):
    """
    Application service container root object
    """

    def __init__(self, name=None, application_settings=None, global_settings=None):
        RootService.__init__(self, settings=global_settings)

        # Compute name for Twisted service
        self.name = u'app-{name}'.format(name=name)

        # Make channel object from application settings configuration object
        application_settings = deepcopy(application_settings)
        application_settings.update(name=name)
        self.channel = Bunch(**application_settings)

        # Create application service object composed of subsystem components
        service = FirmwareBuilderService(
            channel = self.channel
        )

        # Register service component with its container
        self.registerService(service)


class FirmwareBuilderService(MultiService, MultiServiceMixin):

    def __init__(self, channel=None):

        MultiService.__init__(self)

        self.channel = channel or Bunch(realm=None, subscriptions=[])

        self.name = u'service-fb-' + self.channel.get('realm', unicode(id(self)))

        # A bunch of locks for constraining concurrent
        # make processes on the same directory.
        self.locks = defaultdict(threading.RLock)

    def setupService(self):
        #self.log(log.info, u'Setting up')
        self.settings = self.parent.settings

        log.info('Providing firmware for {repository} at {source}', source=self.channel.source, repository=self.channel.repository)
        self.setupSource()

    def setupSource(self):
        # Configure data source

        # There should be just a single instance of a HTTP server service object
        self.source_service = HttpServerService.create(self.settings)

        # Wrap source channel settings into address object
        self.source_address = ForwarderAddress(self.channel.source)

        # Register URI route representing source channel
        self.source_service.registerEndpoint(
            methods     = self.source_address.predicates,
            path        = self.source_address.uri.path,
            callback    = self.build)

    def build(self, bucket):

        # 1. Map/transform topology address information

        # Transformation dict
        data = {}

        # Update transformation dict with information from url capturing
        data.update(bucket.tdata.copy())

        # Update transformation dict with information from request (body/params)
        data.update(bucket.data)

        # Automatically derive "MQTT_TOPIC" from telemetry channel address information
        if 'MQTT_TOPIC' not in data:
            if 'TELEMETRY_REALM' in data and data['TELEMETRY_REALM']:
                data['MQTT_TOPIC'] = '{TELEMETRY_REALM}/{TELEMETRY_USER}/{TELEMETRY_SITE}/{TELEMETRY_NODE}/data.json'.format(**data)

        # Extract option fields and delete them from transformation dict
        options = Bunch()
        option_fields = ['url', 'ref', 'update_submodules', 'path', 'architecture', 'makefile', 'suffix']
        for field in option_fields:
            if field in data:
                options[field] = data[field]
                del data[field]

        # Sanity checks
        # TODO: Check "options.suffix" for being "hex" or "elf" only

        # Override url option with repo url from settings
        options.url = self.channel.repository

        # Propagate path to ESP_ROOT
        # TODO: Clone from https://github.com/esp8266/Arduino
        options.esp_root = self.channel.esp_root

        # Further "options" manipulations

        # Type coercion with default value
        if 'update_submodules' in options:
            options['update_submodules'] = asbool(options['update_submodules'])
        else:
            options['update_submodules'] = True


        # 2. Reporting
        log.info('Building firmware, options={options}, data={data}', options=dict(options), data=dict(data))


        # 3. Build firmware, capturing error output, if any
        payload = self.build_firmware_safe(bucket=bucket, options=bunchify(options), data=bunchify(data))


        # 4. TODO: Announce to appropriate channel via MQTT

        return payload

    def build_firmware_safe(self, options=None, **kwargs):
        # As a very rough guideline, we take a lock on the parameter obtained as working dir
        # to have exclusively one compiler / make process running in the very same directory.
        with self.locks[options.path]:
            return self.build_firmware(options=options, **kwargs)

    def build_firmware(self, bucket=None, options=None, data=None):

        # FirmwareBuilder machinery
        fwbuilder = FirmwareBuilder(
            repo_url=options.url, repo_branch=options.ref, update_submodules=options.update_submodules,
            workingdir=options.path,
            architecture = options.get('architecture'), esp_root=options.esp_root)

        # Run build process with capturing its output
        with fwbuilder.capture() as result:

            # Acquire source code from git repository
            fwbuilder.acquire_source()

            # Patch files with individual parameters
            """
            data = {
                'BOARD_TAG': 'mega2560',
                'BOARD_SUB': 'atmega2560',
                'HE_USER': 'Hotzenplotz',
                'HE_SITE': 'Buxtehude',
                'HE_HIVE': 'Raeuberhoehle',
                'GPRSBEE_APN': 'internet.altes-land.de',
                'GPRSBEE_VCC': 23,
                }
            """
            patch_files = read_list(self.channel.patch_files)
            fwbuilder.patch_files(patch_files, data)

            # Run build process
            fwbuilder.run_build(makefile=options.makefile)
            artefact = fwbuilder.make_artefact()

        if result.success:

            # Compute extended firmware name suffix from transformation dict
            # TODO: Route this information to "Artefact" object
            """
            fwparams = ''
            if data:
                data = data.copy()
                del data['slot']
                dlist = []
                for key, value in data.iteritems():
                    dlist.append(u'{key}={value}'.format(**locals()))
                fwparams = u'-' + u','.join(dlist)
            """

            filename = '{realm}_{name}.{suffix}'.format(
                realm=self.channel.realm, name=artefact.fullname, suffix=options.suffix)

            log.info(u'Build succeeded, filename: {filename}, artefact: {artefact}', filename=filename, artefact=artefact)

            bucket.request.setHeader('Content-Type', 'application/octet-stream')
            bucket.request.setHeader('Content-Disposition', 'attachment; filename={filename}'.format(filename=filename))

            # "options.suffix" may be "hex" or "elf"
            payload = artefact.get_binary(options.suffix)
            return payload

        else:
            error_message = fwbuilder.error_message()
            log.error(u'Build failed\n{message}', message=error_message)
            bucket.request.setResponseCode(http.BAD_REQUEST)
            bucket.request.setHeader('Content-Type', 'text/plain; charset=utf-8')
            return error_message.encode('utf-8')

    def startService(self):
        self.setupService()
        self.log(log.info, u'Starting')
        MultiService.startService(self)

    def log(self, level, prefix):
        level('{prefix} {class_name}. name={name}, channel={channel}',
            prefix=prefix, class_name=self.__class__.__name__, name=self.name, channel=dict(self.channel))


def boot(name=None, **kwargs):
    app = FirmwareBuilderApplication(name=name, **kwargs)
    return app




# -*- coding: utf-8 -*-
# (c) 2014 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from kotori.io.node.bricks import BinaryInputPort, BinaryOutputPort, TimedBinarySemaphore, Blinker, BinaryTopicSignal

PirMotionDetector = BinaryInputPort
PrivacyButton = BinaryInputPort
SignalLight = BinaryOutputPort
OperatorPresenceIndicator = BinaryTopicSignal


class FeatureBase(object):

    # port allocation
    PORT_LED = 'P8_13'
    PORT_PRIVACY_BUTTON = 'P8_15'
    PORT_PIR_SENSOR = 'P8_19'

    def __init__(self, node_id, bus, event_filters=None):
        self.node_id = node_id
        self.bus = bus
        self.event_filters = event_filters
        self.state = None

    @staticmethod
    def state_for_display(state):
        return state and 'enabled' or 'disabled'

    def run_filters(self, state):
        if self.event_filters:
            for event_filter in self.event_filters:
                state = event_filter(state)
        return state

    def publish(self, name, data):
        payload = {'node_id': self.node_id}
        payload.update(data)
        self.bus.publish(name, payload)


class ActivityMonitor(FeatureBase):

    def start(self, holdtime):
        """
        Feature for detecting activity of arbitrary creatures and reacting on that.

        - When triggered:
            - Turn LED into slow blinking mode
            - Publish "activity(True)" message to bus network
        - After "holdtime":
            - Turn LED off
            - Publish "activity(False)" message to bus network
        """

        # define a slow blinking light (default interval 0.5s)
        self.blinking_light = Blinker(SignalLight(self.PORT_LED))
        PirMotionDetector(self.PORT_PIR_SENSOR, signal=TimedBinarySemaphore(holdtime=holdtime, callback=self.on_event))

    def on_event(self, state, *args, **kwargs):

        if state == self.state: return
        self.state = state

        state = self.run_filters(state)

        print "INFO: Activity is", self.state_for_display(state)

        # turn led on/off
        if state:
            self.blinking_light.set()
        else:
            self.blinking_light.unset()

        # publish activity to bus network
        self.publish('broadcast:node-activity', {'state': state})


class PrivacyMonitor(FeatureBase):

    def start(self, holdtime):
        """
        Feature for disabling motion detection for some time, if user hits privacy button.

        - When triggered:
            - Turn LED on
            - Publish "privacy(True)" message to bus network
            - Store privacy state internally (for disabling "monitor_activity")
        - After "holdtime":
            - Turn LED off
            - Publish "privacy(False)" message to bus network
        """

        # define a basic light
        self.light = SignalLight(self.PORT_LED)
        PrivacyButton(self.PORT_PRIVACY_BUTTON, signal=TimedBinarySemaphore(holdtime=holdtime, callback=self.on_event))


    def on_event(self, state, *args, **kwargs):

        if state == self.state: return
        self.state = state

        state = self.run_filters(state)

        print "INFO: Privacy is", self.state_for_display(state)

        # turn led on/off and set/clear internal privacy state
        if state:
            self.light.set()
        else:
            self.light.unset()

        # publish privacy to bus network
        self.publish('broadcast:node-privacy', {'state': state})


class FeatureSet(FeatureBase):

    def __init__(self, node_id, bus):
        FeatureBase.__init__(self, node_id, bus)

        self.privacy_enabled = False

        self.privacy_monitor = PrivacyMonitor(self.node_id, self.bus, event_filters=[self.privacy_filter])
        self.activity_monitor = ActivityMonitor(self.node_id, self.bus, event_filters=[self.activity_filter])

        try:
            self.privacy_monitor.start(holdtime=60)
            self.activity_monitor.start(holdtime=5)
            self.show_operator_presence()
        except Exception as ex:
            print "ERROR: Feature start failed:", ex


    def privacy_filter(self, state):
        self.privacy_enabled = state
        self.activity_monitor.on_event(False)
        return state

    def activity_filter(self, state):
        # skip further actions if privacy mode is enabled
        if state and self.privacy_enabled:
            print "INFO: Will signal no activity due to privacy mode being enabled"
            state = False
        return state

    def show_operator_presence(self):
        """
        Feature for showing whether there's an operator present.

        - When idle:
            - Listens to "broadcast:operator-presence" messages on bus network

        - When triggered:
            - True:  Turn LED into fast blinking mode
            - False: Turn LED off
        """

        # define a fast blinking light (default interval 0.1s)
        blinking_light = Blinker(SignalLight(self.PORT_LED), interval=0.1)

        # connect bus message to blinking light
        OperatorPresenceIndicator(self.bus, 'broadcast:operator-presence', blinking_light)



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import os
from appdirs import user_cache_dir
from kotori.daq.intercom.c import LibraryAdapter, StructRegistryByID
#from kotori.daq.intercom.cffi_adapter import LibraryAdapterCFFI

def setup_h2m_structs_pyclibrary():
    cache_dir = os.path.join(user_cache_dir('kotori'), 'lst')
    if not os.path.isdir(cache_dir): os.makedirs(cache_dir)
    lib_dir = os.path.join(os.path.dirname(__file__), 'cpp')
    library = LibraryAdapter(u'h2m_structs.h', u'h2m_structs.so', include_path=lib_dir, library_path=lib_dir, cache_path=cache_dir)
    struct_registry = StructRegistryByID(library)
    return struct_registry

def setup_h2m_structs_cffi():
    cache_dir = os.path.join(user_cache_dir('kotori'), 'lst')
    if not os.path.isdir(cache_dir): os.makedirs(cache_dir)
    lib_dir = os.path.join(os.path.dirname(__file__), 'cpp')
    library = LibraryAdapterCFFI(u'h2m_structs.h', u'h2m_structs.so', include_path=lib_dir, library_path=lib_dir, cache_path=cache_dir)
    struct_registry = StructRegistryByID(library)
    return struct_registry

setup_h2m_structs = setup_h2m_structs_pyclibrary



from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

# deprecated docopt definition
"""
  #kotori master [--debug]
  #kotori node --master=<> [--debug]
"""

# legacy bootstrap code from iLaundry
"""
from kotori.io.master.server import boot_master
from kotori.io.node.nodeservice import boot_node

# defaults
if config.has_section('wamp'):
    websocket_uri = unicode(config.get('wamp', 'listen'))

# run master and web gui
if 'master' in options:
    boot_master(websocket_uri, debug)
    #boot_web(35000, '', debug)
    #boot_udp_adapter(udp_port, debug)

# run node and web gui only, using a remote master
if 'node' in options:
    websocket_uri = options['--master']
    #boot_web(35000, websocket_uri, debug)
    boot_node(websocket_uri, debug)
    #boot_udp_adapter(udp_port, debug)

# run master, node and web gui
else:
    boot_master(websocket_uri, debug=debug)
    boot_web(http_port, websocket_uri, debug=debug)
    boot_frontend(frontend_port, websocket_uri, debug=debug)
    boot_node(websocket_uri, debug=debug)
    boot_udp_adapter(udp_port, debug=debug)
    boot_sql_database(websocket_uri)
    boot_mongo_database(websocket_uri)
    boot_influx_database(websocket_uri)
"""



# -*- coding: utf-8 -*-
# (c) 2015-2016 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from bunch import Bunch
from twisted.logger import Logger
from kotori.daq.services import RootService
from kotori.daq.services.mig import MqttInfluxGrafanaService
from kotori.daq.intercom.strategies import WanBusStrategy
from kotori.daq.graphing.grafana.manager import GrafanaManager

log = Logger()

class MqttKitApplication(RootService):
    """
    Application service container root object
    """

    def __init__(self, name=None, application_settings=None, global_settings=None):
        RootService.__init__(self, settings=global_settings)

        # Compute name for Twisted service
        self.name = u'app-{name}'.format(name=name)

        # Make channel object from application settings configuration object
        self.channel = Bunch(**application_settings)

        # Create application service object composed of subsystem components
        service = MqttInfluxGrafanaService(
            channel  = self.channel,
            # Data processing strategy and graphing components
            strategy = WanBusStrategy(),
            graphing = GrafanaManager(settings=global_settings, channel=self.channel)
            )

        # Register service component with its container
        self.registerService(service)


def mqttkit_application(name=None, **kwargs):
    app = MqttKitApplication(name=name, **kwargs)
    return app



# -*- coding: utf-8 -*-
# (c) 2015-2016 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from twisted.logger import Logger
from twisted.python.failure import Failure
from autobahn.twisted.wamp import ApplicationRunner

log = Logger()

class WampBus(object):

    def __init__(self, uri=None, application_factory=None, session_factory=None):
        self.uri = uri
        self.application_factory = application_factory or self.application_factory
        self.session_factory = session_factory

    def connect(self):

        # Connect to crossbar router/broker
        self.runner = self.application_factory()

        # Pass session factory to Wamp application runner
        d = self.runner.run(self.session_factory, start_reactor=False)
        d.addErrback(self.on_error)

    def application_factory(self):
        # TODO: Obtain the realm from outside
        return ApplicationRunner(self.uri, u'kotori-realm')

    def on_error(self, ex, *args, **kwargs):
        log.failure(u'Problem starting {name}, please also check if WAMP broker '
                    u'"crossbar" is running at {uri}. args={args!s}, kwargs={kwargs!s}', failure=ex,
            name=self.__class__.__name__, uri=self.uri, args=args, kwargs=kwargs)


class WampSessionMixin(object):
    def onUserError(self, ex, msg):
        """
        Override of wamp.ApplicationSession
        """
        # Handle exception and print to the logs
        log.failure(msg, failure=Failure(ex))



# -*- coding: utf-8 -*-
# (c) 2014 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from kotori.io.node.gpio import GpioInput, GpioOutput

class BinaryInputPort(object):

    def __init__(self, portname, signal=None):
        self.portname = portname
        self.signal = signal

        # detect edge on port
        self.gpio = GpioInput(self.portname, self.sensor_on)

    def sensor_on(self, port):
        print "DEBUG: Port {0} EVENT".format(self.portname)

        # signal sensor-on
        if self.signal:
            self.signal.set(True)

    # currently defunct
    """
    def sensor_off(self, port):
        print "SENSOR OFF:", port

        # signal sensor-off
        if self.callback:
            self.callback(False)
    """


class BinaryOutputPort(object):

    # TODO: handle different ports, not only GPIO

    #HIGH = 1
    #LOW = 2
    #BLINK = 3

    def __init__(self, portname):
        self.portname = portname
        #self.flavor = flavor or self.HIGH
        self.gpio = GpioOutput(self.portname)

        #self.modifier = modifier

    def set(self, *args, **kwargs):
        print "DEBUG: Port {0} ON".format(self.portname)
        self.gpio.on()

        # signal gpio port
        #if self.flavor == self.HIGH:
        #elif self.flavor == self.BLINK:

    def unset(self):
        print "DEBUG: Port {0} OFF".format(self.portname)

        #self.transition_stop()

        # signal gpio port
        self.gpio.off()



class Blinker(object):

    def __init__(self, port, interval=0.5):
        self.port = port
        self.interval = interval
        self.state = None
        self.transition = None

    def set(self, *args, **kwargs):
        self.blink()

    def unset(self):
        self.transition_stop()
        self.port.unset()

    def blink(self):
        self.transition_stop()
        self.transition = LoopingCall(self.blink_task)
        self.transition.start(self.interval)

    def transition_stop(self):
        if self.transition and self.transition.running:
            self.transition.stop()

    def blink_task(self):
        if self.state:
            self.state = False
            self.port.set()
        else:
            self.state = True
            self.port.unset()


class TimedBinarySemaphore(object):

    def __init__(self, holdtime=None, callback=None):
        self.holdtime = holdtime
        self.callback = callback
        self.timer = None

    def set(self, *args, **kwargs):
        #print "SEMAPHORE SET:  ", self.port

        # signal actor-on
        if self.callback:
            self.callback(True, *args, **kwargs)

        # start/reset timer for motion-off signal
        if self.holdtime:
            #print "SEMAPHORE SET:  ", self.port, "holding for {0} seconds".format(self.holdtime)
            if self.timer and not self.timer.called:
                self.timer.cancel()
            self.timer = reactor.callLater(self.holdtime, self.unset)

    def unset(self, *args, **kwargs):
        #print "SEMAPHORE UNSET:", self.port

        # signal sensor-off
        if self.callback:
            self.callback(False, *args, **kwargs)


class BinaryTopicSignal(object):
    """Receives a signal from bus network and signals receiver (actor/port) accordingly"""

    def __init__(self, bus, topic, signal):
        self.bus = bus
        self.topic = topic
        self.signal = signal
        self.bus.subscribe(self.topic, self.process_event)

    def process_event(self, topic, state):
        if state:
            self.signal.set()
        else:
            self.signal.unset()



# -*- coding: utf-8 -*-
# (c) 2015-2016 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import os
from bunch import Bunch
from binascii import hexlify
from copy import deepcopy
from pkg_resources import resource_filename
from twisted.logger import Logger
from twisted.internet import reactor
from kotori.util.configuration import read_list
from kotori.daq.intercom.c import LibraryAdapter, StructRegistryByID
from kotori.daq.intercom.wamp import WampApplication, WampSession
from kotori.daq.intercom.udp import UdpBusForwarder
from kotori.daq.storage.influx import BusInfluxForwarder
from kotori.daq.graphing.grafana import GrafanaManager
from kotori.vendor.lst.message import BinaryMessageAdapter

log = Logger()

class UDPReceiver(object):
    """
    - Receive binary messages via UDP
    - Decode using information from ``h2m_struct.h``
    - Publish decoded messages to software bus in appropriate format
    """

    def __init__(self, bus, config):
        """
        - Initialize UDP receiver with software bus and configuration
        - Create decoder object from appropriate C/C++ header file
        """
        self.bus = bus
        self.config = config

        udp_port = int(self.config['_active_']['udp_port'])
        log.info('Starting UDPReceiver on port {}'.format(udp_port))

        # create decoder adapter
        self.messenger = setup_binary_message_adapter(self.config)

        self.run()

    def run(self):
        """
        Run forwarding component, which actually moves data between UDP and the software bus
        """
        udp_port = int(self.config['_active_']['udp_port'])
        topic = unicode(self.config['_active_']['wamp_topic'])

        log.info('Starting udp adapter on port "{}", publishing to topic "{}"'.format(udp_port, topic))
        reactor.listenUDP(
            udp_port,
            UdpBusForwarder(
                bus=self.bus,
                topic=topic,
                transform=[self.binary_decode, self.bus_encode],
                logger=log))

    def binary_decode(self, payload):
        """
        Decode binary message with information from C/C++ header file
        """
        struct = self.messenger.decode(payload)
        #self.messenger.pprint(struct)
        return struct

    def bus_encode(self, struct):
        """
        Mungle decoded message before publishing to software bus, add two more fields:
        - ``_name_``: name of the struct
        - ``_hex_``:  raw message payload encoded in hex
        """

        # plain
        #data = self.messenger.to_dict(struct)

        # with transformation rules
        data = self.messenger.transform(struct)

        data['_name_'] = struct._name_()
        data['_hex_'] = hexlify(struct._dump_())
        # bus message should be a list of tuples to keep field order
        data_bus = data.items()
        return data_bus


class InfluxStorage(BusInfluxForwarder):
    """
    Receive messages from software bus and store them into InfluxDB timeseries database
    """

    # TODO: Improve parameter passing
    def __init__(self, *args, **kwargs):
        BusInfluxForwarder.__init__(self, *args, **kwargs)

        # grafana setup
        try:
            self.graphing = GrafanaManager(settings=self.config, channel=self.channel)
        except Exception as ex:
            log.failure('Error starting GrafanaManager')

    def storage_location(self, data):
        """
        Compute storage location (database- and timeseries-names) from message information
        """
        sanitize = self.sanitize_db_identifier
        location = Bunch({
            'database': sanitize(self.topic),
            'series':   '{:02d}_{}'.format(data['ID'], sanitize(data['_name_'].replace('struct_', ''))),
        })
        return location

    def store_encode(self, data):
        """
        Filter designated fields from message payload before storing into database
        """
        blacklist_full  = ['_name_', '_hex_', 'ID', 'length', 'ck']
        blacklist_medium  = ['_name_', 'ID', 'length', 'ck']
        blacklist_minor = ['_name_', 'ID']
        blacklist = blacklist_medium
        for entry in blacklist:
            if entry in data:
                del data[entry]
        return data

    def on_store(self, location, data):
        # provision graphing subsystem
        try:
            self.graphing.provision(location, data)
        except Exception:
            log.failure(u'Failed provisioning Grafana')


class StorageAdapter(object):
    """
    Encapsulation for receiving messages from software bus, mungling and storing into InfluxDB
    """

    def __init__(self, bus, config):
        self.bus = bus
        self.config = config
        log.info('Starting InfluxStorage')
        # TODO: Refactor the _active_ mechanics
        topic = unicode(self.config['_active_']['wamp_topic'])
        channel = Bunch(
            settings = self.config['_active_']
        )
        # TODO: Improve parameter passing
        InfluxStorage(bus=self.bus, topic=topic, config=self.config, channel=channel)


class UdpSession(WampSession):
    component = UDPReceiver

class StorageSession(WampSession):
    component = StorageAdapter


def setup_binary_message_adapter(config):
    # TODO: refactor towards OO; e.g. BinaryMessageAdapterFactory

    # TODO: improve path handling
    include_path = config['_active_']['include_path']
    if not include_path.startswith('/'):
        kotori_path = resource_filename('kotori', '')
        include_path = os.path.join(kotori_path, '..', include_path)

    # build and load library
    library = LibraryAdapter.from_header(
        include_path=include_path,
        header_files=config['_active_']['header_files'].split(','))

    # register structs
    struct_registry = StructRegistryByID(library)

    # wrap into decoder adapter
    adapter = BinaryMessageAdapter(struct_registry=struct_registry)

    return adapter


def lst_boot(config, debug=False):
    # TODO: refactor towards OO; e.g. BinaryMessageApplicationFactory

    wamp_uri = unicode(config.wamp.uri)

    # activate/mount multiple "lst" applications
    for channel_name in read_list(config['lst']['channels']):
        log.info('Starting "lst" channel "{}"'.format(channel_name))
        channel = config[channel_name]

        config = deepcopy(config)
        config['_active_'] = channel

        WampApplication(url=wamp_uri, realm=u'lst', session_class=UdpSession, config=config).make()
        WampApplication(url=wamp_uri, realm=u'lst', session_class=StorageSession, config=config).make()



# -*- coding: utf-8 -*-
# (c) 2015-2018 Andreas Motl, <andreas@getkotori.org>
import os
import json

import attr
from bunch import Bunch
from jinja2 import Template
from pkg_resources import resource_filename
from twisted.logger import Logger
from pyramid.settings import asbool

log = Logger()

@attr.s
class GrafanaDashboardModel(object):
    name = attr.ib()
    title = attr.ib()
    datasource = attr.ib()
    measurement_sensors = attr.ib()
    measurement_events = attr.ib()
    uid = attr.ib(default=None)

@attr.s
class GrafanaDashboardBuilder(object):

    grafana_api = attr.ib()
    channel = attr.ib()
    topology = attr.ib()
    model = attr.ib()

    def make(self, data=None):

        dashboard_uid = self.model.uid
        dashboard_name = self.model.name
        dashboard_title = self.model.title
        datasource = self.model.datasource

        # Get or create Grafana folder for stuffing all instant dashboards into
        folder = self.grafana_api.ensure_instant_folder()
        folder_id = folder and folder.get('id') or None

        # Get dashboard if already exists
        dashboard_data = self.grafana_api.get_dashboard(name=dashboard_name)

        # Debugging
        #print 'dashboard_data:'; pprint(dashboard_data)

        # Wrap everything into convenience object
        dashboard = GrafanaDashboard(
            channel=self.channel,
            uid=dashboard_uid,
            title=dashboard_title,
            datasource=datasource,
            folder_id=folder_id,
            dashboard_data=dashboard_data)

        # Generate panels
        panels_new = self.panel_generator(data=data)
        #print 'panels_new:'; pprint(panels_new)

        # Create whole dashboard with all panels
        if not dashboard.dashboard_data:

            # Compute title for dashboard row
            row_title = self.row_title()

            # Build dashboard representation
            dashboard.build(measurement=self.model.measurement_sensors, row_title=row_title, panels=panels_new)

            # Optionally, add annotations
            dashboard.update_annotations(measurement=self.model.measurement_events)

            # Create dashboard
            self.grafana_api.create_dashboard(dashboard, name=dashboard_name)

        else:

            # Update existing dashboard with new annotations
            #annotations = dashboard_data.get('annotations', {}).get('list', [])
            dashboard.update_annotations(measurement=self.model.measurement_events)

            # Update existing dashboard with new panels
            if self.provision_new_panels(dashboard, panels_new):

                # Update dashboard with new panels
                self.grafana_api.create_dashboard(dashboard, name=dashboard_name)

            # Update existing panel with new targets
            else:
                title = self.panel_title()
                panel = self.find_panel_by_title(dashboard_data, title)
                panel_new = self.find_new_panel_by_title(panels_new, title)

                # Find existing targets (field names)
                existing_fields = []
                for target in panel.get('targets', []):
                    existing_fields.append(target.get('alias'))

                new_fields = list(set(panel_new.get('fieldnames', [])) - set(existing_fields))

                if new_fields:

                    for new_field in sorted(new_fields):
                        new_target = dashboard.get_target(panel=panel_new, measurement=self.model.measurement_sensors, fieldname=new_field)
                        new_target_json = json.loads(dashboard.build_target(new_target))
                        panel['targets'].append(new_target_json)

                    # Update dashboard with new panel targets
                    self.grafana_api.create_dashboard(dashboard, name=dashboard_name)

    def find_panel_by_title(self, dashboard, title):
        #print 'find panel by title:', title
        schema_version = dashboard.get('schemaVersion', 6)

        if schema_version >= 16:
            rows = [dashboard]
        elif schema_version >= 6:
            rows = dashboard.get('rows', [])

        for row in rows:
            for panel in row.get('panels', []):
                if panel.get('title') == title:
                    return panel
        return {}

    def find_new_panel_by_title(self, panels_new, title):
        for panel in panels_new:
            if panel.get('title') == title:
                return panel
        return {}

    def provision_new_panels(self, dashboard, panels_new):

        schema_version = dashboard.dashboard_data.get('schemaVersion', 6)

        # Compute which panels are missing
        # TODO: this is hardcoded on row=0
        if schema_version >= 16:
            panels_exists = dashboard.dashboard_data['panels']
        elif schema_version >= 6:
            panels_exists = dashboard.dashboard_data['rows'][0]['panels']
        panels_exists_titles = [panel['title'] for panel in panels_exists]
        panels_new_titles = [panel['title'] for panel in panels_new]

        # v1 - naive
        #panels_missing_titles = set(panels_new_titles) - set(panels_exists_titles)

        # v2 - prefix search
        panels_missing_titles = []
        for new_title in panels_new_titles:
            found = False
            for existing_title in panels_exists_titles:
                exact_match = existing_title == new_title
                fuzzy_match = existing_title.startswith(new_title + ' ')
                if exact_match or fuzzy_match:
                    found = True
                    break
            if not found:
                panels_missing_titles.append(new_title)

        log.debug(u'\n' + \
                    u'Actual titles: {panels_exists_titles},\n' + \
                    u'Target panels: {panels_new},\n' + \
                    u'Target titles: {panels_new_titles}',
            panels_exists_titles=panels_exists_titles,
            panels_new=panels_new,
            panels_new_titles=panels_new_titles,
        )

        if panels_missing_titles:

            log.info(u'Adding missing panels {panels_missing_titles}', panels_missing_titles=panels_missing_titles)

            # establish new panels
            for panel in panels_new:
                panel_title = panel.get('title')
                if panel_title in panels_missing_titles:
                    panels_exists.append(dashboard.build_panel(panel=panel, measurement=self.model.measurement_sensors))

            return True

        else:
            log.info('No missing panels to add')

    def row_title(self):
        return self.topology.network

    def panel_title_human(self):
        parts = []
        if 'node' in self.topology:
            parts.append('device={node}')
        if 'gateway' in self.topology:
            parts.append('site={gateway}')
        return ', '.join(parts).format(**self.topology)

    def panel_title(self, fieldname_prefixes=None):
        """
        return u'{measurement} @ {suffix}'.format(
            measurement = self.model.measurement_sensors,
            suffix = self.panel_title_suffix())
        """
        #fieldname_prefixes = fieldname_prefixes or []
        title = self.panel_title_human()
        if not title:
            title = self.model.measurement_sensors

        title_prefix = self.panel_title_prefix(fieldname_prefixes)
        if title_prefix:
            title = u'{prefix} @ {title}'.format(prefix=title_prefix, title=title)

        return title

    def panel_title_prefix(self, fieldname):
        return

    def panel_generator(self, data):
        #print 'panel_generator, measurement: {}, data: {}'.format(measurement, data)

        # Generate single panel
        panel = self.get_panel_data(data)

        return [panel]

    def get_panel_data(self, data, title=None, fieldname_prefixes=None):

        #fieldname_prefixes = fieldname_prefixes or []

        # Compute title
        if not title:
            title = self.panel_title(fieldname_prefixes)

        # Compute field names
        fieldnames = self.collect_fields(data, fieldname_prefixes)

        # Compute tags
        #tags = self.get_panel_tags()

        panel = {
            'title': title,
            'fieldnames': fieldnames,
            #'tags': tags,
        }

        panel.update(self.get_panel_options(data, fieldname_prefixes))

        return panel

    def get_panel_options(self, data, fieldname_prefixes):
        return {}

    """
    def get_panel_tags(self):
        tags = OrderedDict()
        if 'gateway' in storage_location:
            tags['gateway'] = storage_location.gateway
        if 'node' in storage_location:
            tags['node']    = storage_location.node
        return tags
    """

    @staticmethod
    def collect_fields(data, prefixes=None, sorted=True):
        """
        Field name collection helper.
        Does a prefix search over all fields in "data" and builds
        a list of field names like temp1, temp2, etc. in sorted order.
        """

        # Filter blacklist fields
        # _hex_ is from intercom.c
        # time is from  intercom.mqtt
        blacklist = ['_hex_', 'time']

        fields = []
        for field in data.keys():
            if field in blacklist:
                continue

            if prefixes is None:
                fields.append(field)

            elif isinstance(prefixes, types.ListType):
                for prefix in prefixes:
                    if field.startswith(prefix) or field.endswith(prefix):
                        fields.append(field)
                        break

        if sorted:
            fields.sort()
        return fields


class GrafanaDashboard(object):

    def __init__(self, channel=None, uid=None, title='default', datasource='default', folder_id=None, dashboard_data=None):
        self.channel = channel or Bunch()
        self.dashboard_uid = uid
        self.dashboard_title = title
        self.datasource = datasource
        self.folder_id = folder_id

        self.dashboard_data = dashboard_data

        # Bookkeeping: Which panel id to use when adding new panels?
        # If a dashboard already exists, use the maximum id currently exists as new index.
        self.panel_id = 0
        if dashboard_data:

            # Debugging dashboard JSON
            """
            msg = '''
            -----------------
            title: {title}
            datasource: {datasource}
            dashboard_data:
            {dashboard_data}
            '''
            log.info(msg, title=title, datasource=datasource, dashboard_data=pformat(dashboard_data))
            """

            schema_version = dashboard_data.get('schemaVersion', 6)
            if schema_version >= 16:

                # FIXME: This is hardcoded on row=0
                if dashboard_data['panels'] == [None]:
                    del dashboard_data['panels'][0]

                panels = dashboard_data['panels']
                panel_ids = [panel['id'] for panel in panels]
                if panel_ids:
                    self.panel_id = max(panel_ids)

            elif schema_version >= 6:

                # FIXME: This is hardcoded on row=0
                if dashboard_data['rows'][0]['panels'] == [None]:
                    del dashboard_data['rows'][0]['panels'][0]

                panels = dashboard_data['rows'][0]['panels']
                panel_ids = [panel['id'] for panel in panels]
                if panel_ids:
                    self.panel_id = max(panel_ids)


        self.tpl_dashboard = self.get_template('grafana-dashboard.json')
        self.tpl_annotation = self.get_template('grafana-annotation.json')
        self.tpl_panel = self.get_template('grafana-panel.json')
        self.tpl_target = self.get_template('grafana-target.json')

    def get_template(self, filename):
        filename = os.path.join('resources', filename)
        return Template(file(resource_filename('kotori.daq.graphing.grafana', filename)).read())

    def build(self, measurement, row_title='default', panels=None):

        panels = panels or []
        #pprint(dict(self.channel))

        dashboard_tags = ['instant']
        if 'realm' in self.channel:
            dashboard_tags.append(self.channel.realm)

        data_dashboard = {
            'id': 'null',
            'uid': self.dashboard_uid,
            'title': self.dashboard_title,
            'row_title': row_title,
            'row_height': '300px',
            'tags': json.dumps(dashboard_tags),
            'datasource': self.datasource,
        }

        # build panels list
        panels_json = []
        for panel in panels:
            panel_json = json.dumps(self.build_panel(panel, measurement))
            panels_json.append(panel_json)

        dashboard_json = self.tpl_dashboard.render(data_dashboard, panels=',\n'.join(panels_json))

        self.dashboard_data = json.loads(dashboard_json)
        #pprint(self.dashboard)
        #raise SystemExit

        return self.dashboard_data

    def build_annotation(self, datasource, measurement):
        data_annotation = {
            'datasource': datasource,
            'name': self.get_annotation_name(measurement),
            'measurement': measurement,
            }
        annotation_json = self.tpl_annotation.render(data_annotation)
        return json.loads(annotation_json)

    def get_annotation_name(self, measurement):
        return 'Events {}'.format(measurement.replace('_events', ''))

    def update_annotations(self, measurement):
        annotations = self.dashboard_data.get('annotations', {}).get('list', [])
        found = False
        for annotation in annotations:
            if annotation['name'] == self.get_annotation_name(measurement):
                found = True
        if not found:
            annotations.append(self.build_annotation(self.datasource, measurement))

    def build_panel(self, panel, measurement):

        self.panel_id += 1

        try:
            legend_right_side = asbool(self.channel.settings.graphing_legend_right_side)
        except AttributeError:
            legend_right_side = False

        data_panel = {
            'id': self.panel_id,
            'datasource': self.datasource,
            'panel_title': panel.get('title', 'default'),
            'left_log_base': panel.get('scale', 1),
            'label_y': panel.get('label', ''),
            'format_y': panel.get('format', 'none'),
            'legend_right_side': json.dumps(legend_right_side),
        }

        # Build targets list from fieldnames
        targets_list = []
        for fieldname in panel['fieldnames']:
            data_target = self.get_target(panel, measurement, fieldname)
            target_json = self.build_target(data_target)
            targets_list.append(target_json)

        targets_json = ',\n'.join(targets_list)
        panel_json = self.tpl_panel.render(data_panel, targets=targets_json)
        #print 'panel_json:', panel_json
        try:
            return json.loads(panel_json)
        except Exception:
            log.failure(u'Failed building valid JSON for Grafana panel. data={data}, json={json}',
                data=data_panel, json=panel_json)

    def get_target(self, panel, measurement, fieldname):
        data_target = {
            'measurement': measurement,
            'name': fieldname,
            'alias': fieldname,
            'tags': json.dumps(self.serialize_tags(panel.get('tags', {}))),
        }
        return data_target

    def build_target(self, data_target):
        return self.tpl_target.render(data_target)

    def serialize_tags(self, tags):
        # Compute tags for WHERE constraint
        """
        "tags": [
                {
                "key": "gateway",
                "operator": "=",
                "value": "area_42"
            },
                {
                "condition": "AND",
                "key": "node",
                "operator": "=",
                "value": "node_1"
            }
        ],
        """
        tag_list = []
        for key, value in tags.iteritems():
            tag_entry = {
                "condition": "AND",
                "key": key,
                "operator": "=",
                "value": value,
            }
            tag_list.append(tag_entry)
        return tag_list

    def get_title(self):
        return self.dashboard_data.get('title')

    def wrap_api(self):
        payload = {
            "dashboard": self.dashboard_data,
            "overwrite": False,
        }
        if self.folder_id is not None:
            payload['folderId'] = self.folder_id
        return payload

    @staticmethod
    def setdefaults(adict, bdict):
        for key, value in bdict.iteritems():
            adict.setdefault(key, value)



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
# https://pypi.python.org/pypi/twisted-mqtt
# https://github.com/astrorafael/twisted-mqtt/
from __future__ import absolute_import
from twisted.logger import Logger
from twisted.internet import reactor
from twisted.application.service import Service
from twisted.internet.endpoints import TCP4ClientEndpoint
from kotori.daq.intercom.mqtt.base import BaseMqttAdapter
from mqtt.client.factory import MQTTFactory

log = Logger()

class TwistedMqttAdapter(BaseMqttAdapter, Service):

    def connect(self):
        log.info('Connecting')
        factory = MQTTFactory(profile=MQTTFactory.PUBLISHER | MQTTFactory.SUBSCRIBER)
        point   = TCP4ClientEndpoint(reactor, self.broker_host, self.broker_port)
        d = point.connect(factory).addCallback(self.gotProtocol)
        d.addErrback(self.on_error)

    def gotProtocol(self, p):
        log.info('gotProtocol, connecting {name}', name=self.name)
        self.protocol = p
        #def later():
        d = p.connect(self.name, keepalive=0, cleanStart=True)
        d.addCallback(self.subscribe)
        #d.addCallback(self.prepareToPublish)
        #reactor.callLater(random.randint(2, 7), later)
        #reactor.callInThread(later)


    def subscribe(self, *args):
        #d = self.protocol.subscribe("foo/bar/baz", 0)
        log.info(u"Subscribing to topics {subscriptions}. protocol={protocol}", subscriptions=self.subscriptions, protocol=self.protocol)
        for topic in self.subscriptions:
            log.info(u"Subscribing to topic '{topic}'", topic=topic)
            # Topic name **must not** be unicode, so casting to string
            e = self.protocol.subscribe(str(topic), 0)

        log.info(u"Setting callback handler: {callback}", callback=self.callback)
        self.protocol.setPublishHandler(self.on_message_twisted)
        """
        def cb(*args, **kwargs):
            log.info('publishHandler got called: name={name}, args={args}, kwargs={kwargs}', name=self.name, args=args, kwargs=kwargs)
            return reactor.callFromThread(self.callback, *args, **kwargs)
        self.protocol.setPublishHandler(cb)
        """


    def on_message_twisted(self, topic, payload, *args):
        # former def on_message(self, topic, payload, qos, dup, retain, msgId):
        kwargs = dict(zip(['qos', 'dup', 'retain', 'msgId'], args))
        log.debug('on_message: name={name}, topic={topic}, payload={payload}, kwargs={kwargs}', name=self.name, topic=topic, payload=payload, kwargs=kwargs)
        return self.callback(topic=topic, payload=payload, **kwargs)



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, <andreas@getkotori.org>
from __future__ import absolute_import

from twisted.logger import (
    LogLevel, globalLogBeginner,
    FileLogObserver, FilteringLogObserver, LogLevelFilterPredicate,
    formatTime, timeFormatRFC3339, formatEvent)

# from mqtt.logger

predicate = None
def startLogging(settings, stream=None, level=LogLevel.debug):
    global predicate

    fileObserver = logObserver(stream)
    predicate    = LogLevelFilterPredicate(defaultLogLevel=level)

    if settings.options.debug_mqtt:
        predicate.setLogLevelForNamespace('kotori.daq.services.mig', LogLevel.debug)
        predicate.setLogLevelForNamespace('kotori.daq.application.mqttkit', LogLevel.debug)

    if settings.options.debug_mqtt_driver:
        predicate.setLogLevelForNamespace('kotori.daq.intercom.mqtt', LogLevel.debug)
        predicate.setLogLevelForNamespace('mqtt', LogLevel.debug)
        predicate.setLogLevelForNamespace('paho.mqtt', LogLevel.debug)
    else:
        predicate.setLogLevelForNamespace('kotori.daq.intercom.mqtt', LogLevel.info)
        predicate.setLogLevelForNamespace('mqtt', LogLevel.info)
        predicate.setLogLevelForNamespace('paho.mqtt', LogLevel.info)

    if settings.options.debug_influx:
        predicate.setLogLevelForNamespace('kotori.daq.storage.influx', LogLevel.debug)

    if settings.options.debug_io:
        predicate.setLogLevelForNamespace('kotori.io', LogLevel.debug)

    observers    = [ FilteringLogObserver(observer=fileObserver, predicates=[predicate]) ]
    globalLogBeginner.beginLoggingTo(observers)

def changeLogLevel(namespace, loglevel=LogLevel.info):
    predicate.setLogLevelForNamespace(namespace, loglevel)


# overwritten from twisted.logger

def logObserver(outFile, timeFormat=timeFormatRFC3339):
    def formatEvent(event):
        return formatLogEvent(
            event, formatTime=lambda e: formatTime(e, timeFormat)
        )

    return FileLogObserver(outFile, formatEvent)


def formatLogEvent(event, formatTime=formatTime):
    eventText = formatEvent(event)
    if not eventText:
        return None

    eventText = eventText.replace(u"\n", u"\n\t")
    timeStamp = formatTime(event.get("log_time", None))

    system = event.get("log_system", None)

    if system is None:
        level = event.get("log_level", None)
        if level is None:
            levelName = u"-"
        else:
            levelName = level.name

        system = u"{namespace}".format(
            namespace=event.get("log_namespace", u"-"),
        )
    else:
        try:
            system = unicode(system)
        except Exception:
            system = u"UNFORMATTABLE"

    return u"{timeStamp} [{system}] {level}: {event}\n".format(
        timeStamp=timeStamp,
        system=system.ljust(35),
        level=levelName.upper().ljust(8),
        event=eventText,
    )


__all__ = [startLogging]



# -*- coding: utf-8 -*-
# (c) 2016 Andreas Motl <andreas.motl@elmyra.de>
import tempfile
from pprint import pprint
from twisted.logger import Logger, LogLevel
from twisted.web.template import renderElement
from kotori.io.export.html import DatatablesPage
from kotori.io.protocol.util import get_data_uri
from kotori.io.export.util import dataframe_index_and_sort

log = Logger()

try:
    import pandas
except ImportError:
    log.failure('Tabular export not available, please install "pandas".', level=LogLevel.warn)

class UniversalTabularExporter(object):
    """
    Universal exporter for tabular data.

    Render pandas DataFrame to Excel (XLSX), HDF5
    and NetCDF formats and as DataTables HTML widget.
    """

    def __init__(self, bucket, dataframe):
        self.bucket = bucket
        self.request = bucket.request
        self.dataframe = dataframe

    def render(self, format, kind=None, buffer=None):

        # Variable aliases
        bucket = self.bucket
        df = self.dataframe

        # Compute group name for HDF5 and NetCDF formats
        # TODO: Optionally prefix with "realm" from "bucket.tdata"
        group_name = bucket.title.short

        if format == 'xlsx':
            # http://pandas.pydata.org/pandas-docs/stable/io.html#io-excel-writer
            # https://stackoverflow.com/questions/28058563/write-to-stringio-object-using-pandas-excelwriter
            with pandas.ExcelWriter('temp.xlsx', engine='xlsxwriter') as excel_writer:
                excel_writer.book.filename = buffer
                df.to_excel(excel_writer, sheet_name=bucket.title.compact[:31], index=False)

        elif format in ['hdf', 'hdf5', 'h5']:

            # Create index from "time" column
            df = dataframe_index_and_sort(df, 'time')

            # http://pandas.pydata.org/pandas-docs/stable/io.html#hdf5-pytables
            t = tempfile.NamedTemporaryFile(suffix='.hdf5')
            try:
                df.to_hdf(t.name, group_name, format='table', data_columns=True, index=False)
                buffer.write(t.read())
            except Exception as ex:
                return self.request.error_response(bucket, with_traceback=True)

        elif format in ['nc', 'cdf']:

            # Create index from "time" column
            df = dataframe_index_and_sort(df, 'time')

            # http://xarray.pydata.org/
            # http://xarray.pydata.org/en/stable/io.html#netcdf
            t = tempfile.NamedTemporaryFile(suffix='.nc')
            try:
                #df.to_xarray().to_netcdf(path=t.name, group=group_name)
                #df.to_xarray().to_netcdf(path=t.name, format='NETCDF4', engine='h5netcdf', group=group_name)
                df.to_xarray().to_netcdf(path=t.name, format='NETCDF4', engine='netcdf4', group=group_name)
                buffer.write(t.read())
            except Exception as ex:
                return self.request.error_response(bucket, with_traceback=True)

        elif format in ['dt', 'datatables']:
            # https://datatables.net/

            # Compute data_uri, forward "from" and "to" parameters
            data_uri = get_data_uri(bucket, 'data.html')

            # Render HTML snippet containing DataTable widget
            page = DatatablesPage(data_uri=data_uri, bucket=bucket)
            bucket.request.setHeader('Content-Type', 'text/html; charset=utf-8')
            return renderElement(bucket.request, page)

